import copy
import functools
import numpy as np
import warnings
import xarray as xr

from ..const import v_x, v_y, s2n, corr
from .. import helpers


commondoc = """
    Returns
    -------
    mask : xr.DataArray
        mask applicable to input dataset with ``ds.velocimetry.filter(mask)``.
        If ``inplace=True``, the dataset will be returned masked with ``mask``.

"""


def _base_mask(time_allowed=False, time_required=False):
    """
    wrapper generator for creating generalized structure masking methods for velocimetry

    Parameters
    ----------
    time_allowed : bool, optional
        If set, the dimension "time" is allowed, if not set, mask method can only be applied on datasets without "time"
    time_required
        If set, the dimension "time" is required, if not set, mask method does not require dimension "time" in dataset.

    Returns
    -------
    func : function
        masking method, decorated with standard procedures
    """
    def decorator_func(mask_func):
        mask_func.__doc__ = f"{mask_func.__doc__}{commondoc}"
        # wrap function so that it takes over the docstring and is seen as integral part of the class
        @functools.wraps(mask_func)
        def wrapper_func(ref, inplace=False, reduce_time=False, *args, **kwargs):
            # check if obj seems to contain velocimetry data
            if reduce_time and "time" in ref._obj:
                ds = ref._obj.mean(dim="time", keep_attrs=True)
            else:
                ds = ref._obj
            if not ds.velocimetry.is_velocimetry:
                raise AssertionError("Dataset is not a valid velocimetry dataset")
            if time_required:
                # then automatically time is also allowed
                if not("time" in ds.dims):
                    raise AssertionError(
                f'This mask requires dimension "time". The dataset does not contain dimension "time" or you have set'
                f'reduce_time=True. Apply this mask without applying any reducers in time.'
            )
            if time_required:
                if not("time" in ds):
                    raise AssertionError(
                f'This mask requires dimension "time". The dataset does not contain dimension "time".'
                f'Apply this mask before applying any reducers in time.'
            )
            if not(time_allowed or time_required) and "time" in ds:
                # function must be applied per time step
                mask = ds.groupby("time", squeeze=False).map(mask_func, **kwargs)
            else:
                # apply the wrapped mask function as is
                mask = mask_func(ds, **kwargs)
            # apply mask if inplace
            if inplace:
                # set the _obj data points
                for var in ref._obj.data_vars:
                    ref._obj[var] = ref._obj[var].where(mask)
            return mask
        return wrapper_func
    return decorator_func

class _Velocimetry_MaskMethods:
    """
    Enables use of ``ds.velocimetry.filter`` functions as attributes on a Dataset containing velocimetry results.
    For example, ``Dataset.velocimetry.filter.minmax``. This will return either the dataset with filtered data using
    For example, ``Dataset.velocimetry.filter.minmax``. This will return either the dataset with filtered data using
    the ``minmax`` filter when ``inplace=True`` or the mask that should be applied to filter when ``inplace=False``
    (default). ds.velocimetry.filter([mask1, mask2, ...]) applies the provided filters in the list of filters on
    the dataset by first combining all masks into one, and then applying that mask on the dataset
    """
    def __init__(self, velocimetry):
        # make the original dataset also available on the plotting object
        self.velocimetry = velocimetry
        self._obj = velocimetry._obj
        # Add to class _FilterMethods

    def __call__(self, mask, inplace=False, *args, **kwargs):
        """
        Parameters
        ----------
        mask : xr.DataArray or list of xr.DataArrays
            mask(s) to be applied on dataset, can have mix of y, x and time y, x dimensions
        *args :
        **kwargs :

        Returns
        -------
        ds : xr.Dataset
            Dataset containing filtered velocimetry results
        """
        if not(isinstance(mask, list)):
            # combine masks
            mask = [mask]
        if inplace:
            for m in mask:
                self._obj[v_x] = self._obj[v_x].where(m)
                self._obj[v_y] = self._obj[v_y].where(m)
                self._obj[corr] = self._obj[corr].where(m)
                self._obj[s2n] = self._obj[s2n].where(m)
        else:
            ds = copy.deepcopy(self._obj)
            for m in mask:
                ds[v_x] = ds[v_x].where(m)
                ds[v_y] = ds[v_y].where(m)
                ds[corr] = ds[corr].where(m)
                ds[s2n] = ds[s2n].where(m)
            return ds
    @_base_mask(time_allowed=True)
    def minmax(self, s_min=0.1, s_max=5.):
        """
    Masks values if the velocity scalar lies outside a user-defined valid range.

    Parameters
    ----------
    s_min : float, optional
        minimum scalar velocity [m s-1] (default: 0.1)
    s_max : float, optional
        maximum scalar velocity [m s-1] (default: 5.)

        """
        s = (self[v_x] ** 2 + self[v_y] ** 2) ** 0.5
        # create filter
        mask = (s > s_min) & (s < s_max)
        return mask

    @_base_mask(time_allowed=True)
    def angle(self, angle_expected=0.5 * np.pi, angle_tolerance=0.25 * np.pi):
        """
    filters on the expected angle. The function filters points entirely where the mean angle over time
    deviates more than input parameter angle_bounds (in radians). The function also filters individual
    estimates in time, in case the user wants this (filter_per_timestep=True), in case the angle on
    a specific time step deviates more than the defined amount from the average.
    note: this function does not work appropriately, if the expected angle (+/- anglebounds) are within
    range of zero, as zero is the same as 2*pi. This exception may be resolved in the future if necessary.

    Parameters
    ----------
    angle_expected : float
        angle (0 - 2*pi), measured clock-wise from vertical upwards direction, expected
        in the velocities, default: 0.5*np.pi (meaning from left to right in the x, y coordinate system)
    angle_tolerance : float (0 - 2*pi)
        maximum deviation from expected angle allowed (default: 0.25 * np.pi).
        """
        angle = np.arctan2(self[v_x], self[v_y])
        mask = np.abs(angle - angle_expected) < angle_tolerance
        return mask

    @_base_mask(time_required=True)
    def count(self, tolerance=0.33):
        """
    Masks locations with a too low amount of valid velocities in time, measured by fraction with ``tolerance``.
    Usually applied *after* having applied several other filters.

    Parameters
    ----------
    tolerance : float (0-1)
        tolerance for fractional amount of valid velocities after all filters. If less than the fraction is
        available, the entire velocity will be set to missings.
        """
        mask = self[v_x].count(dim="time") > tolerance * len(self.time)
        return mask


    @_base_mask(time_allowed=True)
    def corr(self, tolerance=0.1):
        """
    Masks values with too low correlation

    Parameters
    ----------
    tolerance : float (0-1)
        tolerance for correlation value (default: 0.1). If correlation is lower than tolerance, it is masked
        """
        return self[corr] > tolerance


    @_base_mask(time_required=True)
    def outliers(self, tolerance=1., mode="or"):
        """
    Mask outliers measured by amount of standard deviations from the mean.

    Parameters
    ----------
    tolerance :  float
        amount of standard deviations allowed from the mean
    mode : str
         can be "and" or "or" (default). If "or" ("and"), then only one (both) of two vector components need(s) to
         be within tolerance.
        """

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            x_std = self[v_x].std(dim="time")
            y_std = self[v_y].std(dim="time")
            x_mean = self[v_x].mean(dim="time")
            y_mean = self[v_y].mean(dim="time")
            x_condition = np.abs((self[v_x] - x_mean) / x_std) < tolerance
            y_condition = np.abs((self[v_y] - y_mean) / y_std) < tolerance
        if mode == "or":
            mask = x_condition | y_condition
        else:
            mask = x_condition & y_condition
        return mask

    @_base_mask(time_required=True)
    def variance(self, tolerance=5, mode="and"):
        """
    Masks locations if their variance (std/mean in time) is above a tolerance level for either or both
    x and y direction.

    Parameters
    ----------
    tolerance :  float
        amount of standard deviations allowed from the mean
    mode : str
         can be "and" (default) or "or". If "or" ("and"), then only one (both) of two vector components need(s) to
         be within tolerance.
        """
        x_std = self[v_x].std(dim="time")
        y_std = self[v_y].std(dim="time")
        x_mean = np.maximum(self[v_x].mean(dim="time"), 1e30)
        y_mean = np.maximum(self[v_y].mean(dim="time"), 1e30)
        x_var = np.abs(x_std / x_mean)
        y_var = np.abs(y_std / y_mean)
        x_condition = x_var < tolerance
        y_condition = y_var < tolerance
        if mode == "or":
            mask = x_condition | y_condition
        else:
            mask = x_condition & y_condition
        return mask


    @_base_mask(time_required=True)
    def rolling(self, wdw=5, tolerance=0.5):
        """
    Masks values if neighbours over a certain rolling length before and after, have a
    significantly higher velocity than value under consideration, measured by tolerance.

    Parameters
    ----------
    wdw : int, optional
        amount of time steps in rolling window (centred) (default: 5)
        """
        s = (self[v_x] ** 2 + self[v_y] ** 2) ** 0.5
        s_rolling = s.fillna(0.).rolling(time=wdw, center=True).max()
        mask = s > tolerance * s_rolling
        return mask


    @_base_mask()
    def window_nan(self, tolerance=0.7, wdw=1, **kwargs):
        """
    Masks values if their surrounding neighbours (inc. value itself) contain too many NaNs. Meant to remove isolated
    velocity estimates.

    Parameters
    ----------
    tolerance : float, optional
        minimum amount of valid values in search window measured as a fraction of total amount of values [0-1]
        (default: 0.3)
    wdw : int, optional
        window size to use for sampling neighbours. zero means, only cell itself, 1 means 3x3 window.
        (default: 1) wdw is used to fill wdw_x_min and wdwd_y_min with its negative (-wdw) value, and wdw_y_min and
        wdw_y_max with its positive value, to create a sampling window.
    wdw_x_min : int, optional
        window size in negative x-direction of grid (must be negative), overrules wdw in negative x-direction if set
    wdw_x_max : int, optional
        window size in positive x-direction of grid, overrules wdw in positive x-direction if set
    wdw_y_min : int, optional
        window size in negative y-direction of grid (must be negative), overrules wdw in negative y-direction if set
    wdw_y_max : int, optional
        window size in positive y-direction of grid, overrules wdw in positive x-direction if set.
        """
        # collect points within a stride, collate and analyze for nan fraction
        ds_wdw = helpers.stack_window(self, wdw=wdw, **kwargs)
        valid_neighbours = ds_wdw[v_x].count(dim="stride")
        mask = valid_neighbours >= tolerance * len(ds_wdw.stride)
        return mask

    @_base_mask()
    def window_mean(self, tolerance=0.7, wdw=1, mode="or", **kwargs):
        """
    Masks values when their value deviates more than tolerance (measured as relative fraction) from the mean of its
    neighbours (inc. itself).

    Parameters
    ----------
    tolerance: float, optional
        amount of velocity relative to the mean velocity  (default: 0.7)
    wdw : int, optional
        window used to determine relevant neighbours
    wdw_x_min : int, optional
        window size in negative x-direction of grid (must be negative), overrules wdw in negative x-direction if set
    wdw_x_max : int, optional
        window size in positive x-direction of grid, overrules wdw in positive x-direction if set
    wdw_y_min : int, optional
        window size in negative y-direction of grid (must be negative), overrules wdw in negative y-direction if set
    wdw_y_max : int, optional
        window size in positive y-direction of grid, overrules wdw in positive x-direction if set.
    mode : str
         can be "and" (default) or "or". If "or" ("and"), then only one (both) of two vector components need(s) to
         be within tolerance.
        """
        # collect points within a stride, collate and analyze for median value and deviation
        ds_wdw = helpers.stack_window(self, wdw=wdw, **kwargs)
        ds_mean = ds_wdw.mean(dim="stride")
        x_condition = np.abs(self[v_x] - ds_mean[v_x]) / ds_mean[v_x] < tolerance
        y_condition = np.abs(self[v_y] - ds_mean[v_y]) / ds_mean[v_y] < tolerance
        if mode == "or":
            mask = x_condition | y_condition
        else:
            mask = x_condition & y_condition
        return mask


    @_base_mask()
    def window_replace(self, wdw=1, iter=1, **kwargs):
        """
    Replaces values in a certain window size with mean of their neighbours. Returns a Dataset instead of a mask.
    NOTE: This functionality may be moved to a different subclass in later releases.

    Parameters
    ----------
    wdw : int, optional
        window used to determine relevant neighbours
        """
        ds = copy.deepcopy(self)
        for n in range(iter):
            # collect points within a stride, collate and analyze for median value and deviation
            ds_wdw = helpers.stack_window(ds, wdw=wdw, **kwargs)
            ds_mean = ds_wdw.mean(dim="stride")
            ds = ds.fillna(ds_mean)
        return ds



"""Video module for pyorc."""

import copy
import json
import os
import warnings
from typing import List, Literal, Optional, Union

import cv2
import dask
import dask.array as da
import numpy as np
import xarray as xr

from pyorc import const, cv, helpers

from .cameraconfig import CameraConfig, get_camera_config, load_camera_config


class Video:  # (cv2.VideoCapture)
    """Video class for reading and extracting data from video files."""

    def __repr__(self):
        template = """
Filename: {:s}
FPS: {:f}
start frame: {:d}
end frame: {:d}
Camera configuration: {:s}
        """.format

        return template(
            self.fn,
            self.fps,
            self.start_frame,
            self.end_frame,
            self.camera_config.__repr__() if hasattr(self, "camera_config") else "none",
        )

    def __init__(
        self,
        fn: str,
        camera_config: Optional[Union[str, CameraConfig]] = None,
        h_a: Optional[float] = None,
        start_frame: Optional[int] = None,
        end_frame: Optional[int] = None,
        freq: int = 1,
        chunksize: int = 20,
        stabilize: Optional[List[List]] = None,
        lazy: bool = True,
        rotation: Optional[int] = None,
        fps: Optional[float] = None,
        progress: bool = True,
    ):
        """Video class, inheriting parts from cv2.VideoCapture.

        Contains a camera configuration to it, and a start and end frame to read from the video. Several methods read
        frames into memory or into a xr.DataArray with attributes. These can then be processed with other pyorc API
        functionalities.

        Parameters
        ----------
        fn : str
            Locally stored video file
        camera_config : pyorc.CameraConfig, optional
            contains all information about the camera, lens parameters, lens position, ground control points with GPS
            coordinates, and all referencing information (see CameraConfig), needed to reproject frames on a horizontal
            geographically referenced plane.
        h_a : float, optional
            actual height [m], measured in local vertical reference during the video (e.g. a staff gauge in view of
            the camera)
        start_frame : int, optional
            first frame to use in analysis (default: 0)
        end_frame : int, optional
            last frame to use in analysis (if not set, last frame available in video will be used)
        freq : int, optional
            Frequency to read frames with. Default is 1, if set to e.g. 2 only each 2nd frame will be read.
        chunksize : int, optional
            Amount of frames to read in one pass, defaults to 20
        stabilize : list of lists, optional
            set of coordinates, that together encapsulate the polygon that defines the mask, separating land from water.
            The mask is used to select region (on land) for rigid point search for stabilization. If not set, then no
            stabilization will be performed
        lazy : bool, optional
            By default set to True, making frames read in organised chunks. If set to False, video is read in memory
            entirely. Likely, we will deprecate this option in the future as we plan to optimize chunked reading.
            Currently if memory size allows, setting lazy to False results in faster processing.
        rotation : int, optional
            can be 0, 90, 180, 270. If provided, images will be forced to rotate along the provided angle.
        fps : float, optional
            hard set for frames per second. Use this with utmost caution and only when you are confident that the video
            metadata is incorrect.
        progress : bool, optional
            Display progress bar while reading video. Default is True.

        """
        assert isinstance(start_frame, (int, type(None))), 'start_frame must be of type "int"'
        assert isinstance(end_frame, (int, type(None))), 'end_frame must be of type "int"'
        self.feats_stats = None
        self.feats_errs = None
        self.ms = None
        self.mask = None
        self.lazy = lazy
        self.progress = progress
        self.stabilize = stabilize
        if camera_config is not None:
            self.camera_config = camera_config
            # if camera_config is not None:
            # check if h_a is supplied, if so, then also z_0 and h_ref must be available
            if h_a is not None:
                assert isinstance(self.camera_config.gcps["z_0"], float), (
                    "h_a was supplied, but camera config's gcps do not contain z_0, this is needed for dynamic "
                    "reprojection. You can supplying z_0 and h_ref in the camera_config's gcps upon making a camera "
                    "configuration. "
                )
                assert isinstance(self.camera_config.gcps["h_ref"], float), (
                    "h_a was supplied, but camera config's gcps do not contain h_ref, this is needed for dynamic "
                    "reprojection. You must supply z_0 and h_ref in the camera_config's gcps upon making a camera "
                    "configuration. "
                )
                # check if h_a is not very far from h_ref
                if np.abs(h_a - self.camera_config.gcps["h_ref"]) > const.WATER_LEVEL_MAX_DIFF:
                    warnings.warn(
                        f"h_a is more than {const.WATER_LEVEL_MAX_DIFF} meters different from h_ref. You may have "
                        "made a mistake in the h_a definition. Check for instance if your h_a is measured with the same"
                        "datum as h_ref. ",
                        stacklevel=2,
                    )
        if not os.path.isfile(fn):
            raise IOError(
                f"Video file {fn} does not exist. If you are on windows and using \\ as file "
                f"separator ensure that the file string is prepended with an r, e.g. "
                f"r'c:\\video_files\\filename.mp4'"
            )

        cap = cv2.VideoCapture(fn)
        # Force to automatically rotate frames
        cap.set(cv2.CAP_PROP_ORIENTATION_AUTO, 1)
        self.height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        # explicitly open file for reading
        if self.stabilize is not None:
            # set a gridded mask based on the roi points
            self.set_mask_from_exterior(self.stabilize)
        # set end and start frame
        self.frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1
        if start_frame is not None:
            if start_frame > self.frame_count > 0:
                raise ValueError("Start frame is larger than total amount of frames")
        else:
            start_frame = 0
        if end_frame is not None:
            if end_frame < start_frame:
                raise ValueError(f"Start frame {start_frame} is larger than end frame {end_frame}")
            # end frame cannot be larger than total amount of available frames
            end_frame = np.minimum(end_frame, self.frame_count)
        else:
            end_frame = self.frame_count
        self.fps = cap.get(cv2.CAP_PROP_FPS)
        self.rotation = rotation
        # extract times, frame numbers and frames as far as available
        time, frame_number, frames = cv.get_time_frames(
            cap, start_frame, end_frame, lazy=lazy, rotation=self.rotation, method="bgr", fps=fps, progress=progress
        )
        self.frames = frames
        # check if end_frame changed
        if len(frame_number) > 0:
            if frame_number[-1] != end_frame:
                warnings.warn(
                    f"End frame {end_frame} cannot be read from file. End frame is adapted to {frame_number[-1]}",
                    stacklevel=2,
                )
                end_frame = frame_number[-1]
        # apparently we are reading an image instead of a video

        self.end_frame = end_frame
        self.freq = freq
        self.chunksize = chunksize
        self.time = time
        self.frame_number = frame_number
        self.start_frame = start_frame
        if self.stabilize is not None:
            # select the right recipe dependent on the movie being fixed or moving
            self.get_ms(cap)
        if fps is None:
            self.fps = cap.get(cv2.CAP_PROP_FPS)
        else:
            self.fps = fps
        # set other properties
        self.h_a = h_a
        # make camera config part of the video object
        self.fn = fn
        self._stills = {}  # here all stills are stored lazily
        # nothing to be done at this stage, release file for now.
        cap.release()
        del cap

    @property
    def lazy(self):
        """Get lazy flag."""
        return self._lazy

    @lazy.setter
    def lazy(self, lazy):
        self._lazy = lazy

    @property
    def mask(self):
        """Get region mask for stabilization."""
        return self._mask

    @mask.setter
    def mask(self, mask: np.ndarray):
        if mask is None:
            self._mask = None
        else:
            self._mask = mask

    @property
    def camera_config(self):
        """Get `CameraConfig` object attached to `Video` instance."""
        if hasattr(self, "_camera_config"):
            return self._camera_config
        else:
            return None

    @camera_config.setter
    def camera_config(self, camera_config_input: Union[str, dict]):
        """Set camera config as a serializable object from either a filename, json string or a dict."""
        try:
            if isinstance(camera_config_input, str):
                if os.path.isfile(camera_config_input):
                    # assume string is a file
                    self._camera_config = load_camera_config(camera_config_input)
                else:  # Try to read CameraConfig from string
                    self._camera_config = get_camera_config(camera_config_input)
            elif isinstance(camera_config_input, CameraConfig):
                # set CameraConfig as is
                self._camera_config = camera_config_input
            elif isinstance(camera_config_input, dict):
                # Create CameraConfig from dict
                self._camera_config = CameraConfig(**camera_config_input)
        except IOError:
            raise IOError(
                "Could not recognise input as a CameraConfig file, string, dictionary or CameraConfig object."
            )

    @property
    def end_frame(self):
        """:return: int, last frame considered in analysis"""
        return self._end_frame

    @end_frame.setter
    def end_frame(self, end_frame: Optional[int] = None):
        # sometimes last frames are not read by OpenCV, hence we skip the last frame always
        if end_frame is None:
            self._end_frame = self.frame_count - 1
        else:
            self._end_frame = end_frame

    @property
    def freq(self):
        """Get video sampling frequency."""
        return self._freq

    @freq.setter
    def freq(self, freq=1):
        self._freq = freq

    @property
    def progress(self):
        """Get progress property."""
        return self._progress

    @progress.setter
    def progress(self, progress=True):
        self._progress = progress

    @property
    def stabilize(self):
        """Get stabilization region coordinates."""
        if self._stabilize is not None:
            return self._stabilize
        elif hasattr(self, "camera_config"):
            if hasattr(self.camera_config, "stabilize"):
                return self.camera_config.stabilize

    @stabilize.setter
    def stabilize(self, coords: Optional[List[List]] = None):
        self._stabilize = coords

    @property
    def h_a(self):
        """:return: Actual water level [m] during video"""
        return self._h_a

    @h_a.setter
    def h_a(self, h_a: float):
        if h_a is not None:
            assert isinstance(h_a, float), f"The actual water level must be a float, you supplied a {type(h_a)}"
            if h_a < 0:
                warnings.warn(
                    "Water level is negative. This can be correct, but may be unlikely, especially if you use a "
                    "staff gauge.",
                    stacklevel=2,
                )
        self._h_a = h_a

    @property
    def start_frame(self):
        """:return: int, first frame considered in analysis"""
        return self._start_frame

    @start_frame.setter
    def start_frame(self, start_frame: Optional[int] = None):
        if start_frame is None:
            self._start_frame = 0
        else:
            self._start_frame = start_frame

    @property
    def frames(self):
        """Get frames of Video instance."""
        return self._frames

    @frames.setter
    def frames(self, frames: Optional[List] = None):
        self._frames = frames

    @property
    def fps(self):
        """:return: float, frames per second"""
        return self._fps

    @fps.setter
    def fps(self, fps: float):
        if (np.isinf(fps)) or (fps <= 0):
            raise ValueError(f"FPS in video is {fps} which is not a valid value. Repair the video file before use")
        self._fps = fps

    @property
    def corners(self):
        """:return: list of 4 lists (int) with [column, row] locations of area of interest in video objective"""
        return self._corners

    @corners.setter
    def corners(self, corners: List[List]):
        self._corners = corners

    @property
    def rotation(self):
        """Get rotation code."""
        if self._rotation is not None:
            return self._rotation
        elif hasattr(self, "camera_config"):
            if hasattr(self.camera_config, "rotation"):
                return helpers.get_rotation_code(self.camera_config.rotation)

    @rotation.setter
    def rotation(self, rotation_code: int):
        """Set rotation from integer in the form of OpenCV rotation codes."""
        self._rotation = helpers.get_rotation_code(rotation_code)

    def get_frame(
        self, n: int, method: Optional[Literal["grayscale", "rgb", "hsv", "bgr"]] = "grayscale"
    ) -> np.ndarray:
        """Retrieve one frame.

        Parameters
        ----------
        n : int
            frame number to retrieve
        method : str
            can be "rgb", "grayscale", or "hsv", default: "grayscale"

        Returns
        -------
        frame : np.ndarray
            2d array (grayscale) or 3d (rgb/hsv) with frame

        """
        assert n >= 0, "frame number cannot be negative"
        assert (
            n - self.start_frame <= self.end_frame - self.start_frame
        ), "frame number is larger than the difference between the start and end frame "
        # assert (method in ["grayscale", "rgb",
        #                    "hsv"]), f'method must be "grayscale", "rgb" or "hsv", method is "{method}"'
        if not os.path.isfile(self.fn):
            raise IOError(f"Video file {self.fn} does not exist.")
        cap = cv2.VideoCapture(self.fn)
        cap.set(cv2.CAP_PROP_POS_FRAMES, n + self.start_frame)
        ret, img = cv.get_frame(cap, rotation=self.rotation, ms=self.ms[n] if self.ms else None, method=method)
        self.frame_count = n + 1
        cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
        cap.release()
        return img

    def get_frames_chunk(
        self, n_start: int, n_end: int, method: Optional[Literal["grayscale", "rgb", "hsv", "bgr"]] = "grayscale"
    ) -> np.ndarray:
        """Retrieve a chunk of frames in one go.

        Parameters
        ----------
        n_start : int
            frame number to initiate retrieval
        n_end : int
            last frame number of retrieval
        method : str
            can be "rgb", "grayscale", or "hsv", default: "grayscale"

        Returns
        -------
        frame : np.ndarray
            2d array (grayscale) or 3d (rgb/hsv) with frame

        """
        assert n_start >= 0, "frame number cannot be negative"
        assert (
            n_start - self.start_frame <= self.end_frame - self.start_frame
        ), "frame number is larger than the difference between the start and end frame "
        # assert (method in ["grayscale", "rgb",
        #                    "hsv"]), f'method must be "grayscale", "rgb" or "hsv", method is "{method}"'
        if not os.path.isfile(self.fn):
            raise IOError(f"Video file {self.fn} does not exist.")
        cap = cv2.VideoCapture(self.fn)
        cap.set(cv2.CAP_PROP_POS_FRAMES, n_start + self.start_frame)
        imgs = []
        for n in range(n_start, n_end):
            ret, img = cv.get_frame(cap, rotation=self.rotation, ms=self.ms[n] if self.ms else None, method=method)
            self.frame_count = n + 1
            imgs.append(img)
        cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
        cap.release()
        return np.array(imgs)

    def get_frames(self, method: Optional[Literal["grayscale", "rgb", "hsv", "bgr"]] = "grayscale") -> xr.DataArray:
        """Get a xr.DataArray, containing a dask array of frames, from `start_frame` until `end_frame`.

        The xr.DataArray will contain all coordinate variables and attributes, needed for further processing. It
        may be lazy, so that frames can be read at a later stage.

        Parameters
        ----------
        method: str, optional
            method for color scaling, can be "grayscale" (default) or "rgb"

        Returns
        -------
        frames : xr.DataArray
            containing all requested frames

        """
        assert hasattr(
            self, "_camera_config"
        ), "No camera configuration is set, add it to the video using the .camera_config method"
        # camera_config may be altered for the frames object, so copy below
        camera_config = copy.deepcopy(self.camera_config)
        # frames_chunk = 20
        if self.frames is None:
            # a specific method for collecting frames is requested or lazy access is requested.
            # get_frame = dask.delayed(self.get_frame, pure=True)  # Lazy version of get_frame
            get_frames_chunk = dask.delayed(self.get_frames_chunk, pure=True)  # Lazy version of get_frame
            # get all listed frames
            # frames = [get_frame(n=n, method=method) for n, f_number in enumerate(self.frame_number)]
            # derive video shape
            sample = get_frames_chunk(n_start=0, n_end=1, method=method).compute()[0]
            data_array = []
            for n_start in range(0, len(self.frame_number), self.chunksize):
                n_end = np.minimum(n_start + self.chunksize, len(self.frame_number))
                frame_chunk = get_frames_chunk(n_start=n_start, n_end=n_end, method=method)
                shape = (n_end - n_start, *sample.shape)
                data_array.append(da.from_delayed(frame_chunk, dtype=sample.dtype, shape=shape))

            da_stack = da.concatenate(data_array, axis=0)
        else:
            da_stack = self.frames
            # apply stabilisation
            if self.ms is not None:
                # da_stack = np.array([cv.transform(img, m) for img, m in zip(da_stack, self.ms)])
                da_stack = np.array([cv.transform(cv.color_scale(img, method), m) for img, m in zip(da_stack, self.ms)])
            else:
                # only color transform
                da_stack = np.array([cv.color_scale(img, method) for img in da_stack])
            sample = da_stack[0]

        time = (
            np.array(self.time) * 0.001
        )  # measure in seconds to comply with CF conventions # np.arange(len(data_array))*1/self.fps
        # y needs to be flipped up down to match the order of rows followed by coordinate systems (bottom to top)
        y = np.flipud(np.arange(sample.shape[0]))
        x = np.arange(sample.shape[1])
        # perspective column and row coordinate grids
        xp, yp = np.meshgrid(x, y)
        coords = {"time": time, "y": y, "x": x}
        if len(sample.shape) == 3:
            coords["rgb"] = np.array([0, 1, 2])
        # make DataArray dimensions and attributes
        dims = tuple(coords.keys())
        attrs = {
            "camera_shape": str([len(y), len(x)]),
            "camera_config": camera_config.to_json(),
            "h_a": json.dumps(self.h_a),
            "chunksize": self.chunksize,
        }
        frames = xr.DataArray(
            da_stack,
            dims=dims,
            coords=coords,
            attrs=attrs,
        )[:: self.freq]
        frames = frames.chunk({"time": 20})  # set chunks over time dimension
        del coords["time"]
        if len(sample.shape) == 3:
            del coords["rgb"]
        # add coordinate grids (i.e. without time)
        frames = frames.frames.add_xy_coords({"xp": xp, "yp": yp}, coords, const.PERSPECTIVE_ATTRS)
        frames.name = "frames"
        return frames

    def set_mask_from_exterior(self, exterior: List[List]):
        """Prepare a mask grid with 255 outside of the stabilization polygon and 0 inside.

        Parameters
        ----------
        exterior : list of lists
            coordinates defining the polygon for masking

        Returns
        -------
        self.mask : np.ndarray
            mask for stabilization region

        """
        mask_coords = np.array([exterior], dtype=np.int32)
        mask = np.zeros((self.height, self.width), np.uint8)
        mask = cv2.fillPoly(mask, [mask_coords], 255)
        mask[mask == 0] = 1
        mask[mask == 255] = 0
        mask[mask == 1] = 255
        self.mask = mask

    def get_ms(self, cap: cv2.VideoCapture, split: Optional[int] = 2):
        """Get stabilization transforms for each frame based on analysis of stable points outside of water area.

        Parameters
        ----------
        cap : cv2.VideoCapture
            ref to video file

        split : int
            Amount of quadrants to split the area in to search for stabilization points. In each quadrant, an equal
            number of stabilization points will be searched for .This is in order to ensure stabilization uses points
            in all quadrants and prevent overfitting on only one part of the objective.

        """
        self.ms = cv.get_ms_gftt(
            cap,
            start_frame=self.start_frame,
            end_frame=self.end_frame,
            split=split,
            mask=self.mask,
            progress=self.progress,
        )



"""pyorc.cli utilities and interactive views."""

import hashlib
import json
import logging
import os
from typing import Optional

import click
import cv2
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import yaml
from shapely.geometry import Point

import pyorc.api
from pyorc import CameraConfig, Video, cv, helpers, load_camera_config
from pyorc.cli.cli_elements import AoiSelect, GcpSelect, StabilizeSelect


def get_corners_interactive(
    fn,
    gcps,
    crs=None,
    crs_gcps=None,
    frame_sample=0.0,
    camera_matrix=None,
    dist_coeffs=None,
    rotation=None,
    logger=logging,
):
    """Select AOI corners interactively using first selected video frame."""
    vid = Video(fn, start_frame=frame_sample, end_frame=frame_sample + 1, rotation=rotation)
    # get first frame
    frame = vid.get_frame(0, method="rgb")
    src = gcps["src"]
    if crs_gcps is not None:
        dst = helpers.xyz_transform(gcps["dst"], crs_from=crs_gcps, crs_to=4326)
    else:
        dst = gcps["dst"]
    # setup preliminary cam config
    cam_config = CameraConfig(
        height=frame.shape[0],
        width=frame.shape[1],
        gcps=gcps,
        crs=crs,
        camera_matrix=camera_matrix,
        dist_coeffs=dist_coeffs,
        rotation=rotation,
    )
    selector = AoiSelect(frame, src, dst, cam_config, logger=logger)
    # uncomment below to test the interaction, not suitable for automated unit test
    plt.show(block=True)
    return selector.src

    # setup a cam_config without


def get_gcps_interactive(
    fn,
    dst,
    crs=None,
    crs_gcps=None,
    frame_sample=0,
    focal_length=None,
    k1=None,
    k2=None,
    lens_position=None,
    rotation=None,
    logger=logging,
):
    """Select GCP points in interactive display using first selected video frame."""
    vid = Video(fn, start_frame=frame_sample, end_frame=frame_sample + 1, rotation=rotation)
    # get first frame
    frame = vid.get_frame(0, method="rgb")
    # construct camera matrix and distortion coefficients
    # parse items
    camera_matrix, dist_coeffs = parse_lens_params(
        height=frame.shape[0], width=frame.shape[1], focal_length=focal_length, k1=k1, k2=k2
    )

    if crs_gcps is not None:
        dst = helpers.xyz_transform(dst, crs_from=crs_gcps, crs_to=4326)
    selector = GcpSelect(
        frame,
        dst,
        crs=crs,
        camera_matrix=camera_matrix,
        dist_coeffs=dist_coeffs,
        lens_position=lens_position,
        logger=logger,
    )
    plt.show(block=True)
    return selector.src, selector.camera_matrix, selector.dist_coeffs


def get_stabilize_pol(fn, frame_sample=0, rotation=None, logger=logging):
    """Select stabilization region in interactive display on first selected frame of provided video file."""
    vid = Video(fn, start_frame=frame_sample, end_frame=frame_sample + 1, rotation=rotation)
    frame = vid.get_frame(0, method="rgb")
    selector = StabilizeSelect(frame, logger=logger)
    plt.show(block=True)
    return selector.src


def get_file_hash(fn):
    """Turn file content into a SHA-256 hash for cross-comparison."""
    hash256 = hashlib.sha256()
    with open(fn, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            # print(byte_block)
            hash256.update(byte_block)
    return hash256


def get_gcps_optimized_fit(src, dst, height, width, c=2.0, camera_matrix=None, dist_coeffs=None, lens_position=None):
    """Fit intrinsic and extrinsic parameters on provided set of src and dst points."""
    # optimize cam matrix and dist coeffs with provided control points
    if np.array(dst).shape == (4, 2):
        _dst = np.c_[np.array(dst), np.zeros(4)]
    else:
        _dst = np.array(dst)
    print(camera_matrix)
    camera_matrix, dist_coeffs, err = cv.optimize_intrinsic(
        src, _dst, height, width, c=c, lens_position=lens_position, camera_matrix=camera_matrix, dist_coeffs=dist_coeffs
    )
    # once optimized, solve the perspective, and estimate the GCP locations with the perspective rot/trans
    coord_mean = np.array(_dst).mean(axis=0)
    _src = np.float32(src)
    _dst = np.float32(np.array(_dst) - coord_mean)
    success, rvec, tvec = cv2.solvePnP(_dst, _src, camera_matrix, np.array(dist_coeffs))

    # estimate source point location
    src_est, jacobian = cv2.projectPoints(_dst, rvec, tvec, camera_matrix, np.array(dist_coeffs))
    src_est = np.array([list(point[0]) for point in src_est])
    dst_est = cv.unproject_points(_src, _dst[:, -1], rvec, tvec, camera_matrix, dist_coeffs)
    # add mean coordinates to estimated locations
    dst_est = np.array(dst_est)[:, 0 : len(coord_mean)] + coord_mean
    # also reverse rvec and tvec to real-world (rw), add mean coordinate, and reverse again.
    rvec_cam, tvec_cam = cv.pose_world_to_camera(rvec, tvec)
    tvec_cam += coord_mean
    rvec, tvec = cv.pose_world_to_camera(rvec_cam, tvec_cam)
    return src_est, dst_est, camera_matrix, dist_coeffs, rvec, tvec, err


def parse_json(ctx, param, value):
    """Parse json file."""
    if value is None:
        return None
    if os.path.isfile(value):
        with open(value, "r") as f:
            kwargs = json.load(f)
    else:
        if value.strip("{").startswith("'"):
            value = value.replace("'", '"')
        try:
            kwargs = json.loads(value)
        except json.JSONDecodeError:
            raise ValueError(f'Could not decode JSON "{value}"')
    return kwargs


def parse_corners(ctx, param, value):
    """Parse corner [x, y] coordinates into a list of coordinates."""
    if value is None:
        return None
    # check if list contains lists of 2 values
    corners = json.loads(value)
    assert len(corners) == 4, "--corners must contain a list of lists with exactly 4 points"
    for n, val in enumerate(corners):
        assert isinstance(val, list), f"--corners value {n} is not a list {val}"
        assert (
            len(val) == 2
        ), f"--corners value {n} must contain row, column coordinate but consists of {len(val)} numbers"
    return [[int(x), int(y)] for x, y in corners]


def parse_lens_params(
    height: int,
    width: int,
    focal_length: Optional[float] = None,
    k1: Optional[float] = None,
    k2: Optional[float] = None,
):
    """Parse lens parameters to camera matrix and distortion coefficients vector."""
    if focal_length is not None:
        camera_matrix = cv.get_cam_mtx(height, width, c=2.0, focal_length=focal_length)
    else:
        camera_matrix = None
    if k1 is not None or k2 is not None:
        dist_coeffs = cv.DIST_COEFFS.copy()
        if k1 is not None:
            dist_coeffs[0][0] = k1
        if k2 is not None:
            dist_coeffs[1][0] = k2
    else:
        dist_coeffs = None
    return camera_matrix, dist_coeffs


def validate_file(ctx, param, value):
    """Validate existence of file."""
    if value is not None:
        if not (os.path.isfile(value)):
            raise click.FileError(f"{value}")
        return value


def validate_dir(ctx, param, value):
    """Validate existence of directory and create if needed."""
    if not (os.path.isdir(value)):
        os.makedirs(value)
    return value


def validate_rotation(ctx, param, value):
    """Validate rotation value within list of possible values."""
    if value is not None:
        if value not in [90, 180, 270, None]:
            raise click.UsageError("Rotation value must be either 90, 180 or 270")
        return value


def parse_camconfig(ctx, param, camconfig_file):
    """Read and validate cam config file.

    Parameters
    ----------
    ctx : click context
        context
    param : Any
        parameter (not used)
    camconfig_file : str,
       file containing camera configuration

    Returns
    -------
    cam_config: dict
        configuration as dictionary

    """
    # validate if file can be read as camera config object
    camconfig = load_camera_config(camconfig_file)

    # return dict formatted
    return camconfig.to_dict_str()


def parse_recipe(ctx, param, recipe_file):
    """Read and validate entire recipe from top to bottom, add compulsory classes where needed.

    Parameters
    ----------
    ctx : click context
        context
    param : Any
        parameter (not used)
    recipe_file : str,
        file containing .yml formatted recipe

    Returns
    -------
    recipe : dict,
        dictionary with entire recipe for running processing

    """
    with open(recipe_file, "r") as f:
        body = f.read()
    recipe = yaml.load(body, Loader=yaml.FullLoader)
    recipe = validate_recipe(recipe)
    return recipe


def parse_src(ctx, param, value):
    """Parse source (column, row) coordinates."""
    value = parse_json(ctx, param, value)
    if value is not None:
        # check if at least 4 of 2
        assert len(value) >= 4, "--src must contain a list of lists [column, row] with at least 4 points"
        for n, val in enumerate(value):
            assert isinstance(val, list), f"--src value {n} is not a list {val}"
            assert (
                len(val) == 2
            ), f"--src value {n} must contain row, column coordinate but consists of {len(val)} numbers"
    return value


def parse_dst(ctx, param, value):
    """Parse destination (real-world) coordinates."""
    value = parse_json(ctx, param, value)
    value = validate_dst(value)
    return value


def parse_str_num(ctx, param, value):
    """Parse strings to numbers."""
    if value is not None:
        try:
            float(value)
        except Exception:
            return value
        if value.isnumeric():
            return int(value)
        else:
            return float(value)


def parse_cross_section_gdf(ctx, param, value):
    """Check if cross-section can be read. Return file itself."""
    if value is None:
        return None
    try:
        gdf, _ = read_shape_as_gdf(fn=value)  # crs is inside the gdf if available, so leave out
    except Exception:
        raise click.FileError(f"There is a problem with the cross section file {value}")
    return value


def read_shape_as_gdf(fn=None, geojson=None, gdf=None):
    """Read shapefile."""
    if fn is None and geojson is None:
        raise click.UsageError("Either fn or geojson must be provided")
    if geojson:
        if "crs" in geojson:
            crs = geojson["crs"]["properties"]["name"]
        else:
            crs = None
        gdf = gpd.GeoDataFrame().from_features(geojson, crs=crs)
    else:
        gdf = helpers.read_shape_safe_crs(fn)
        crs = gdf.crs
    # check if all geometries are points
    assert all([isinstance(geom, Point) for geom in gdf.geometry]), (
        "shapefile may only contain geometries of type " '"Point"'
    )
    return gdf, crs


def read_shape(fn=None, geojson=None):
    """Read shapefile."""
    gdf, crs = read_shape_as_gdf(fn=fn, geojson=geojson)
    if gdf.geometry[0].has_z:
        coords = [[p.x, p.y, p.z] for p in gdf.geometry]
    else:
        coords = [[p.x, p.y] for p in gdf.geometry]
    if not (hasattr(gdf, "crs")):
        click.echo("shapefile or geojson does not contain CRS, assuming CRS is the same as camera config CRS")
        crs = None
    elif gdf.crs is None:
        click.echo("shapefile or geojson does not contain CRS, assuming CRS is the same as camera config CRS")
        crs = None
    else:
        crs = gdf.crs.to_wkt()
    return coords, crs


def validate_dst(value):
    """Validate destination (real-world) coordinates."""
    if value is not None:
        if len(value) in [2, 4]:
            # assume [x, y] pairs are provided
            len_points = 2
        elif len(value) < 6:
            raise click.UsageError(
                f"--dst must contain exactly 2 or 4 with [x, y], or at least 6 with [x, y, z] points, "
                f"contains {len(value)}."
            )
        else:
            len_points = 3
        for n, val in enumerate(value):
            assert isinstance(val, list), f"--dst value {n} is not a list {val}"
            assert len(val) == len_points, (
                f"--dst value {n} must contain 3 coordinates (x, y, z) "
                f"but consists of {len(val)} numbers, value is {val}"
            )
    return value


def validate_recipe(recipe):
    """Validate recipe."""
    valid_classes = ["video", "water_level", "frames", "velocimetry", "mask", "transect", "plot"]  # allowed classes
    required_classes = ["video", "frames", "velocimetry"]  # mandatory classes (if not present, these are added)
    check_args = {
        "video": "video",
        "frames": "frames",
        # "velocimetry": "frames"
    }  # check if arguments to underlying methods called by recipe section are available and get valid arguments
    # skip_checks = ["plot"]  # these sections are not checked for valid inputs
    process_methods = [
        "write"
    ]  # methods that are specifically needed within process steps and not part of pyorc class methods
    for k in recipe:  # main section
        if k not in valid_classes:
            raise ValueError(f"key '{k}' is not allowed, must be one of {valid_classes}")
        # if k in check_args:
        # loop through all methods and check if their inputs are valid
        for m in recipe[k]:  # method in section
            if recipe[k][m] is None:
                # replace for empty dict
                recipe[k][m] = {}
            if m not in process_methods and k in check_args:
                # get the subclass that is called within the section of interest
                cls = getattr(pyorc.api, check_args[k].capitalize())
                if not hasattr(cls, m):
                    raise ValueError(f"Class '{check_args[k].capitalize()}' does not have a method or property '{m}'")
                method = getattr(cls, m)
                # find valid args to method
                if callable(method):
                    if k in check_args:
                        # check if kwargs is contained in method
                        if "kwargs" in method.__code__.co_varnames:
                            valid_args = None
                        else:
                            valid_args = method.__code__.co_varnames[
                                : method.__code__.co_argcount
                            ]  # get input args to method
                        if valid_args:
                            for arg in recipe[k][m]:
                                if arg not in valid_args:
                                    raise ValueError(
                                        f"Method '{check_args[k].capitalize()}.{m}' does not have input "
                                        f"argument '{arg}', must be one of {valid_args}"
                                    )
    # add empty dicts for missing but compulsory classes
    for _c in required_classes:
        if _c not in recipe:
            # add empties for compulsory recipe components
            recipe[_c] = {}
    return recipe



"""PIV processing wrappers for FF-PIV."""

import gc
import warnings

import numpy as np
import xarray as xr
from ffpiv import cross_corr, u_v_displacement, window
from tqdm import tqdm


def load_frame_chunk(da):
    """Load frame chunk into memory. If not successful try to reduce size of frame chunk."""
    da_loaded = da.copy(deep=True)
    try:
        da_loaded = da.load()
    except TypeError:
        da_loaded = da[:-1]
        da_loaded = load_frame_chunk(da_loaded)
    return da_loaded


def get_ffpiv(
    frames,
    y,
    x,
    dt,
    window_size,
    overlap,
    search_area_size,
    res_y,
    res_x,
    chunksize=None,
    memory_factor=2,
    engine="numba",
):
    """Compute time-resolved Particle Image Velocimetry (PIV) using Fast Fourier Transform (FFT) within FF-PIV.

    This function calculates the velocity field from a sequence of image frames using PIV techniques. The process
    involves dividing the images into interrogation windows, performing cross-correlation analysis to capture
    displacement, and then calculating the velocity components. The function efficiently handles large datasets by
    computing PIV per chunk of frames and manages memory usage.

    The method generally is much faster than OpenPIV because it utilizes parallelized functions implemented in numba.

    Parameters
    ----------
    frames : xr.DataArray
        image frames to be processed (pyorc.Frames compatible).
    y : np.ndarray
        Array of expected y-coordinates of velocimetry fields.
    x : np.ndarray
        Array of x-coordinates.
    dt : array_like
        Array of time intervals between each frame pair.
    window_size : [int, int]
        Size of the interrogation window.
    overlap : [int, int]
        Overlap between interrogation windows.
    search_area_size : [int, int]
        Size of the search area for cross-correlation.
    res_y : float
        Spatial resolution in the y-direction.
    res_x : float
        Spatial resolution in the x-direction.
    chunksize : int, optional
        if provided, the frames will be treated per chunk with `chunksize` as provided, if not, optimal chunk size is
        estimated based on available memory.
    memory_factor : float, optional
        available memory is divided by this factor to estimate the chunk size. Default is 4.
    engine : str, optional
        ff-piv engine to use, can be "numpy" or "numba". "numba" is generally much faster.

    Returns
    -------
    ds : xarray.Dataset
        Dataset containing the signal-to-noise ratio (`s2n`), correlation maxima (`corr`), and velocity
        components (`v_x`, `v_y`).

    """
    CHUNK_SIZE_ERROR = (
        "Chunk size with selected nr of chunks ({chunks}) is 2 or less. If you manually "
        "selected `chunks={chunks}` then consider increasing chunk size to at least 2, and preferrably more. If memory "
        "is limited, consider closing memory intensive applications. If pyorc crashes, then this is due to "
        " insufficient memory."
    )
    CHUNK_SIZE_WARNING = (
        "Memory availability is poor ({avail_mem} GB). Chunk size is automatically set to {chunksize} to avoid "
        "memory issues. If pyorc crashes, then this is due to insufficient memory. Consider to manually set a lower "
        "chunk size e.g using `get_piv(engine={engine}, chunk=2)` or `get_piv(engine={engine}, chunk=3)` or close "
        "memory intensive applications."
    )
    # compute memory availability and size of problem to pipe to ffpiv functions
    dim_size = frames[0].shape
    req_mem = window.required_memory(
        n_frames=len(frames),
        dim_size=dim_size,
        window_size=window_size,
        overlap=overlap,
        search_area_size=search_area_size,
    )
    if chunksize is None:
        # estimate chunk size
        avail_mem = window.available_memory() / memory_factor
        chunks = int((req_mem // avail_mem) + 1)
        chunksize = int(np.ceil((len(frames)) / chunks))
        if chunksize <= 5:
            warnings.warn(
                CHUNK_SIZE_WARNING.format(avail_mem=avail_mem / 1e9, chunksize=chunksize, engine=engine), stacklevel=2
            )
            chunksize = 5  # hard override, try to manage with 5
            chunks = int(np.ceil((len(frames)) / chunksize))
    if chunksize < 2:
        raise OverflowError(CHUNK_SIZE_ERROR.format(chunks=chunks))
    frames_chunks = [frames[np.maximum(chunk * chunksize - 1, 0) : (chunk + 1) * chunksize] for chunk in range(chunks)]
    # check if there are chunks that are too small in size, needs to be at least 2 frames per chunk
    frames_chunks = [frames_chunk for frames_chunk in frames_chunks if len(frames_chunk) >= 2]
    n_rows, n_cols = len(y), len(x)

    # make progress bar
    pbar = tqdm(range(len(frames_chunks)), position=0, leave=True)
    pbar.set_description("Computing PIV per chunk")

    # Loop over list
    ds_piv_chunks = []  # datasets with piv results per chunk
    for n in pbar:
        da = frames_chunks[n]
        # get time slice
        da = load_frame_chunk(da)
        time = da.time[1:]
        dt_chunk = dt.sel(time=time)
        # check length again, only if ge 2, assess velocities
        if len(da) >= 2:
            # perform cross correlation analysis yielding correlations for each interrogation window
            x_, y_, corr = cross_corr(
                da.values,
                window_size=window_size,
                overlap=overlap,
                search_area_size=search_area_size,
                normalize=False,
                engine=engine,
                verbose=False,
            )
            frames_chunks[n] = None
            del da
            gc.collect()

            # get the maximum correlation per interrogation window
            corr_max = np.nanmax(corr, axis=(-1, -2))

            # get signal-to-noise, whilst suppressing nanmean over empty slice warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                s2n = corr_max / np.nanmean(corr, axis=(-1, -2))

            # reshape corr / s2n to the amount of expected rows and columns
            s2n = (s2n.reshape(-1, n_rows, n_cols)).astype(np.float32)
            corr_max = (corr_max.reshape(-1, n_rows, n_cols)).astype(np.float32)
            u, v = u_v_displacement(corr, n_rows, n_cols, engine=engine)
            # convert into meter per second and store as float32 to save memory / disk space
            u = (u * res_x / np.expand_dims(dt_chunk, (1, 2))).astype(np.float32)
            v = (v * res_y / np.expand_dims(dt_chunk, (1, 2))).astype(np.float32)
            # put s2n, corr_max, u and v in one xarray dataset, with coordinates time, y and x
            ds = xr.Dataset(
                {
                    "s2n": (["time", "y", "x"], s2n),
                    "corr": (["time", "y", "x"], corr_max),
                    "v_x": (["time", "y", "x"], u),
                    "v_y": (["time", "y", "x"], v),
                },
                coords={
                    "time": time,
                    "y": y,
                    "x": x,
                },
            )
            # u and v to meter per second
            ds_piv_chunks.append(ds)
    # concatenate all parts in time
    ds = xr.concat(ds_piv_chunks, dim="time")
    return ds



"""API for pyorc."""

from .cameraconfig import CameraConfig, get_camera_config, load_camera_config
from .cross_section import CrossSection
from .frames import Frames
from .transect import Transect
from .velocimetry import Velocimetry
from .video import Video

__all__ = [
    "CameraConfig",
    "CrossSection",
    "load_camera_config",
    "get_camera_config",
    "Video",
    "Frames",
    "Velocimetry",
    "Transect",
]



"""pyorc: free and open-source image-based surface velocity and discharge."""

__version__ = "0.8.9"

from .api import CameraConfig, CrossSection, Frames, Transect, Velocimetry, Video, get_camera_config, load_camera_config  # noqa
from .project import *  # noqa
from . import cli, service, sample_data  # noqa

__all__ = [
    "CameraConfig",
    "load_camera_config",
    "get_camera_config",
    "Video",
    "Frames",
    "Velocimetry",
    "Transect",
    "CrossSection",
    "service",
    "cli",
    "sample_data",
]



"""Service module for pyorc."""

import copy
import functools
import json
import logging
import os.path
import subprocess
from typing import Dict, Optional

import click
import numpy as np
import xarray as xr
import yaml
from dask.diagnostics import ProgressBar
from matplotlib.colors import Normalize

import pyorc
from pyorc import CameraConfig, CrossSection, Video, const
from pyorc.cli import cli_utils
from pyorc.helpers import read_shape_safe_crs

__all__ = ["velocity_flow", "velocity_flow_subprocess"]

logger = logging.getLogger(__name__)


def get_water_level(
    video: Video,
    cross_section: CrossSection,
    n_start: int = 0,
    n_end: int = 1,
    method: const.ALLOWED_COLOR_METHODS_WATER_LEVEL = "grayscale",
    frames_options: Optional[Dict] = None,
    water_level_options: Optional[Dict] = None,
):
    water_level_options = {} if water_level_options is None else water_level_options
    frames_options = {} if frames_options is None else frames_options

    if method not in ["grayscale", "hue", "sat", "val"]:
        raise ValueError(
            f"Method {method} not supported for water level detection, choose one"
            f" of {const.ALLOWED_COLOR_METHODS_WATER_LEVEL}"
        )
    da_frames = video.get_frames(method=method)[n_start:n_end]
    # preprocess
    da_frames = apply_methods(da_frames, "frames", logger=logger, skip_args=["to_video"], **frames_options)
    # if preprocessing still results in a time dim, average in time
    if "time" in da_frames.dims:
        da_mean = da_frames.mean(dim="time")
    else:
        da_mean = da_frames
    # extract the image
    img = np.uint8(da_mean.values)
    h_a = cross_section.detect_water_level(img, **water_level_options)
    return h_a


def vmin_vmax_to_norm(opts):
    """Convert vmin and/or vmax into norm option.

    This is done to ensure plotting options work for all plotting methods.

    Parameters
    ----------
    opts : dict
        Dictionary with kwargs to pass to plotting method of pyorc.

    Returns
    -------
    opts : dict
        Dictionary with vmin/vmax replaced by norm if these are provided by user

    """
    if "vmin" in opts or "vmax" in opts:
        vmin = opts.pop("vmin", None)
        vmax = opts.pop("vmax", None)
        norm = Normalize(vmin=vmin, vmax=vmax)
        opts["norm"] = norm
    return opts


def apply_methods(obj, subclass, logger=logger, skip_args=None, **kwargs):
    if skip_args is None:
        skip_args = []
    for m, _kwargs in kwargs.items():
        if m not in skip_args:
            # get the subclass with the expected methods
            cls = getattr(obj, subclass)
            if not (hasattr(cls, m)):
                raise ValueError(f'Method "{m}" for {subclass} does not exist, please check your recipe')
            logger.debug(f"Applying {m} on {subclass} with parameters {_kwargs}")
            meth = getattr(cls, m)
            obj = meth(**_kwargs)

    return obj


def get_masks(obj, **mask_methods):
    masks = []
    for m, _kwargs in mask_methods.items():
        if _kwargs is None:
            # make an empty dict to pass args
            _kwargs = {}
        meth = getattr(obj.velocimetry.mask, m)
        masks.append(meth(**_kwargs))
    return masks


def run_func_hash_io(
    attrs=None,
    inputs=None,
    configs=None,
    outputs=None,
    write_path=".pyorc",
    check=False,
):
    """Check if inputs to function have changed and or output is present or not.

    Wrapper function to run a sub function if either output is not present or input has changed. If check is False then
    simply passes everything.
    """
    attrs = [] if attrs is None else attrs
    inputs = [] if inputs is None else inputs
    configs = [] if configs is None else configs
    outputs = [] if outputs is None else outputs

    def decorator_func(processor_func):
        @functools.wraps(processor_func)
        def wrapper_func(ref, *args, **kwargs):
            func_name = processor_func.__name__
            # set output path for state files
            path_out = os.path.join(ref.output, write_path)
            if not (os.path.isdir(path_out)):
                os.makedirs(path_out)
            # default, assume running will take place
            run = True
            if check and ref.update:
                run = False  # assume you don't run unless at least one thing changed
                # start with config checks
                fn_recipe = os.path.join(path_out, f"{ref.prefix}{func_name}.yml")
                if not (os.path.isfile(fn_recipe)):
                    run = True
                else:
                    recipe_part = {c: ref.recipe[c] for c in configs if c in ref.recipe}
                    with open(fn_recipe, "r") as f:
                        cfg_ancient = f.read()
                    cfg = yaml.dump(recipe_part, default_flow_style=False, sort_keys=False)
                    if cfg != cfg_ancient:
                        # config has changed
                        ref.logger.debug(f'Configuration of "{func_name}" has changed, requiring rerun')
                        run = True
                if not (run):
                    # if configs are not changed, then go to the file integrity checks
                    for i in inputs + outputs:
                        fn = getattr(ref, i)
                        fn_hash = os.path.join(path_out, f"{os.path.basename(getattr(ref, i))}.hash")
                        if not (os.path.isfile(fn)):
                            run = True
                            break
                        else:
                            # also check if hash file exists
                            if not (os.path.isfile(fn_hash)):
                                run = True
                                break
                            else:
                                hash256 = cli_utils.get_file_hash(fn)
                                with open(fn_hash, "r") as f:
                                    hash256_ancient = f.read()
                                if hash256.hexdigest() != hash256_ancient:
                                    ref.logger.debug(
                                        f"File integrity of {fn} has changed, requiring rerun of {func_name}"
                                    )
                                    run = True
                                    break
            if run:
                # apply the wrapped processor function
                ref.logger.info(f"Running {func_name}")
                processor_func(ref, *args, **kwargs)
                # after run, store configuration file and hashes of in- and outputs
                fn_recipe = os.path.join(path_out, f"{ref.prefix}{func_name}.yml")
                recipe_part = {c: ref.recipe[c] for c in configs if c in ref.recipe}
                with open(fn_recipe, "w") as f:
                    yaml.dump(recipe_part, f, default_flow_style=False, sort_keys=False)
                # after run, store input and output hashes
                for i in inputs + outputs:
                    fn_hash = os.path.join(path_out, f"{os.path.basename(getattr(ref, i))}.hash")
                    # get hash
                    hash256 = cli_utils.get_file_hash(getattr(ref, i))
                    with open(fn_hash, "w") as f:
                        f.write(hash256.hexdigest())
            else:
                ref.logger.info(
                    f'Configuration, dependencies, input and output files for section "{func_name}" have not changed '
                    f"since last run, skipping..."
                )
                for attr, output in zip(attrs, outputs):
                    if attr is not None:
                        fn = getattr(ref, output)
                        ref.logger.info(
                            f'Results for section "{func_name}" already available, reading from {os.path.abspath(fn)}'
                        )
                        setattr(ref, attr, xr.open_dataset(fn))

        return wrapper_func

    return decorator_func


class VelocityFlowProcessor(object):
    """General processor class for processing of videos in to velocities and flow and derivative products."""

    def __init__(
        self,
        recipe: Dict,
        videofile: str,
        cameraconfig: Dict,
        prefix: str,
        output: str,
        h_a: Optional[float] = None,
        cross: Optional[str] = None,
        cross_wl: Optional[str] = None,
        update: bool = False,
        concurrency: bool = True,
        fn_piv: str = "piv.nc",
        fn_piv_mask: str = "piv_mask.nc",
        fn_transect_template: str = "transect_{:s}.nc",
        logger: logging.Logger = logging,
    ):
        """Initialize processor with settings and files.

        Parameters
        ----------
        recipe : dict
            YAML recipe, parsed from CLI
        videofile : str
            path to video
        cameraconfig : dict
            camera config as dict (not yet loaded as CamerConfig object)
        prefix : str
            prefix of produced output files
        output : str
            path to output file
        h_a : float, optional
            Current water level in meters
        cross : str, optional
            path to cross-section coordinates and crs file in GeoJSON or other readible by geopandas
            to be used for discharge estimation.
        cross_wl : str, optional
            path to cross-section coordinates and crs file in GeoJSON or other readible by geopandas
            to be used for water level detection.
        update : bool, optional
            if set, only update components with changed inputs and configurations
        concurrency : bool, optional
            if set to False, then dask will only run synchronous preventing overuse of memory. This will be slower
        fn_piv : str, optional
            Path to piv result file (default "piv.nc")
        fn_piv_mask : str, optional
            Path to piv masked result file (default "piv_mask.nc")
        fn_transect_template : str, optional
            Path template to transect result files (default: "transect_{:s}.nc")
        logger : logging, optional
            reference to logger instance

        """
        logger.debug("Initializing Velocity Flow Processor")
        cross_section_wl = None
        camera_config = CameraConfig(**cameraconfig)
        if h_a is not None:
            if abs(h_a - cameraconfig["gcps"]["h_ref"]) > const.WATER_LEVEL_MAX_DIFF:
                logger.warning(
                    f"Water level is very different from reference. Difference is {h_a - cameraconfig['gcps']['h_ref']}"
                    f" meter. Check if your water level, units and/or datum are correct. This difference may result in "
                    "no projected pixels inside area of interest."
                )
        if h_a is None and recipe["video"].get("h_a") is None and cross is None:
            raise click.UsageError(
                "No actual water level is found, and no water level available in recipe. Either provide a water level,"
                " or provide a valid cross section with `--cross` to estimate water level optically. See"
                " pyorc velocimetry --help for more information."
            )
        if h_a is not None:
            recipe["video"]["h_a"] = h_a
            logger.info(f"Water level provided as argument: h = {h_a} m.")  # cross section wl will NOT be used!
        elif cross_wl is not None:
            logger.info(
                "Cross section for water level detection provided, and no water level set, "
                " water level will be estimated optically."
            )
            gdf = read_shape_safe_crs(cross_wl)
            cross_section_wl = pyorc.CrossSection(camera_config=camera_config, cross_section=gdf)
            if "water_level" not in recipe:
                # make sure water_level is represented
                recipe["water_level"] = {}
        elif recipe["video"].get("h_a") is not None:
            logger.info(f"Water level provided in recipe: h = {recipe['video']['h_a']} m.")
        else:
            logger.error(
                "No water level provided on CLI and no cross section provided. If you only process one video, use the "
                "same value as h_ref in your camera config, by changing your command as follows: "
                f"pyorc velocimetry ... <repeat as before> --h_a {camera_config.gcps['h_ref']}"
            )
            raise click.Abort()
        # check what projection method is used, use throughout
        self.proj_method = "cv"
        proj = recipe["frames"].get("project")
        if proj:
            if proj.get("method") == "numpy":
                self.proj_method = "numpy"
        self.update = update  # set to True when checks are needed if data already exists or not
        self.recipe = recipe
        self.output = output
        self.concurrency = concurrency
        self.prefix = prefix
        # set cross section for water levels
        self.cross_section_wl = cross_section_wl
        # for now also provide a cross section for flow extraction, use the same.
        self.cross_section_fn = cross  # TODO use this property to extract velocity transects.
        self.fn_piv = os.path.join(self.output, prefix + fn_piv)
        self.fn_piv_mask = os.path.join(self.output, prefix + fn_piv_mask) if "mask" in recipe else self.fn_piv
        self.fn_transect_template = (
            os.path.join(self.output, prefix + fn_transect_template).format if "transect" in recipe else None
        )
        if self.fn_transect_template is not None:
            self.fn_transects = [self.fn_transect_template(t) for t in recipe["transect"]]
        self.read = True
        self.write = False
        self.fn_video = videofile
        self.cam_config = camera_config
        self.logger = logger
        # TODO: perform checks, minimum steps required
        self.logger.info("Velocity Flow Processor initialized")

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output):
        self._output = output

    @property
    def read(self):
        return self._read

    @read.setter
    def read(self, read):
        self._read = read

    @property
    def write(self):
        return self._write

    @write.setter
    def write(self, write):
        self._write = write

    @property
    def fn_cam_config(self):
        return self._fn_cam_config

    @fn_cam_config.setter
    def fn_cam_config(self, fn_cam_config):
        self._fn_cam_config = fn_cam_config

    @property
    def fn_video(self):
        return self._fn_video

    @fn_video.setter
    def fn_video(self, fn_video):
        self._fn_video = fn_video

    def process(self):
        """Perform all velocimetry processing in logical pre-defined order.

        The process also checks for compulsory steps.
        """
        self.logger.info("Starting velocimetry processing pipeline")
        if not self.concurrency:
            import dask

            # run only synchronous
            dask.config.set(scheduler="synchronous")
        # dask.config.set(pool=Pool(4))
        # dask.config.set(scheduler='processes')
        self.video(**self.recipe["video"])
        # if cross_section is set, then water levels are missing and must be obtained
        if self.cross_section_wl is not None:
            self.water_level(**self.recipe["water_level"])
        self.frames(**self.recipe["frames"])
        self.velocimetry(**self.recipe["velocimetry"])
        if "mask" in self.recipe:
            self.mask(**self.recipe["mask"])
        else:
            # no masking so use non-masked velocimetry as masked
            self.velocimetry_mask_obj = self.velocimetry_obj
        if "transect" in self.recipe:
            if self.cross_section_fn is not None:
                # ensure that the cross section is available for transect measurements
                self.recipe["transect"]["transect_1"]["shapefile"] = self.cross_section_fn
            self.transect(**self.recipe["transect"])
        if "plot" in self.recipe:
            self.plot(**self.recipe["plot"])
        # remove all potentially memory consumptive attributes
        self.da_frames.close()
        self.velocimetry_mask_obj.close()
        self.velocimetry_obj.close()
        delattr(self, "video_obj")
        delattr(self, "velocimetry_obj")
        delattr(self, "velocimetry_mask_obj")
        delattr(self, "da_frames")
        self.logger.info("Velocimetry processing pipeline completed :-)")
        return None
        # TODO .get_transect and check if it contains data,

        #  perform any post processing such as plotting or possibly later other analyses

    # def wrapper
    # when status is update, then a wrapper function should check the consistency of all input data for each function
    # block with previous runs. If not consistent then rerun. Do this automatically if data is not present or
    # -y is provided, do with user intervention if stale file is present or -y is not provided

    def video(self, **kwargs):
        try:
            self.logger.debug(f"Reading video {self.fn_video} from file")
            self.video_obj = Video(self.fn_video, camera_config=self.cam_config, **kwargs)
            # some checks ...
            self.logger.info(f"Video successfully read from {self.fn_video}")
        except Exception as e:
            self.logger.error(f"Could not read video from {self.fn_video}. Error: {e}")
            raise Exception(f"Could not read video from {self.fn_video}. Error: {e}")

    def water_level(self, **kwargs):
        try:
            self.logger.debug("Estimating water level from video by crossing water line with cross section.")
            h_a = get_water_level(self.video_obj, cross_section=self.cross_section_wl, **kwargs)
            if h_a is None:
                self.logger.error("Water level could not be estimated from video. Please set a water level with --h_a.")
                raise click.Abort()
            # set the found water level on video
            self.logger.info("Water level estimated optically h = {:1.3f} m. in local datum.".format(h_a))
            self.video_obj.h_a = h_a
        except Exception as e:
            self.logger.error(f"Could not estimate water level from video. Error: {e}")
            raise Exception(f"Could not estimate water level from video. Error: {e}")

    def frames(self, **kwargs):
        # start with extracting frames
        try:
            self.logger.debug("Retrieving frames from video.")
            self.da_frames = self.video_obj.get_frames()
            self.logger.debug(f"Retrieved {len(self.da_frames)} from video.")
            if "project" not in kwargs:
                kwargs["project"] = {}
            # iterate over steps in processing
            self.da_frames = apply_methods(
                self.da_frames, "frames", logger=self.logger, skip_args=["to_video"], **kwargs
            )
            if "to_video" in kwargs:
                kwargs_video = kwargs["to_video"]
                self.logger.info(f"Writing video of processed frames to {kwargs_video['fn']}")
                self.da_frames.frames.to_video(**kwargs_video)
            self.logger.info("Frames retrieved and preprocessed.")
        except Exception as e:
            self.logger.error(f"Could not extract frames from video. Error: {e}")
            raise Exception(f"Could not extract frames from video. Error: {e}")

    @run_func_hash_io(
        attrs=["velocimetry_obj"],
        check=True,
        inputs=["fn_video"],
        configs=["video", "frames", "velocimetry"],
        outputs=["fn_piv"],
    )
    def velocimetry(self, method="get_piv", write=False, **kwargs):
        self.logger.debug(f"Performing velocimetry with {method}.")
        try:
            if len(kwargs) > 1:
                raise OverflowError(f"Too many arguments under velocimetry, only one allowed, but {len(kwargs)} given.")
            kwargs[method] = kwargs.get(method, {}) if len(kwargs) == 0 else kwargs[method]
            # get velocimetry results
            self.velocimetry_obj = apply_methods(self.da_frames, "frames", logger=self.logger, **kwargs)
            m = list(kwargs.keys())[0]
            parameters = kwargs[m]
            self.logger.info(f"Velocimetry derived with method {m} with parameters {parameters}")
            if write:
                delayed_obj = self.velocimetry_obj.to_netcdf(self.fn_piv, compute=False)
                with ProgressBar():
                    delayed_obj.compute()
                del delayed_obj
                self.logger.info(f"Velocimetry written to {self.fn_piv}")
                # Load the velocimetry into memory to prevent re-writes in next steps
                delattr(self, "velocimetry_obj")
                self.velocimetry_obj = xr.open_dataset(self.fn_piv)
            self.logger.info("Velocimetry successfully derived.")
        except Exception as e:
            self.logger.error(f"Could not derive velocimetry from frames. Error: {e}")
            raise Exception(f"Could not derive velocimetry from frames. Error: {e}")

    @run_func_hash_io(
        attrs=["velocimetry_mask_obj"],
        check=True,
        inputs=["fn_piv"],
        configs=["video", "frames", "velocimetry", "mask"],
        outputs=["fn_piv_mask"],
    )
    def mask(self, write=False, **kwargs):
        try:
            self.logger.debug("Applying masks to velocimetry.")
            self.velocimetry_mask_obj = copy.deepcopy(self.velocimetry_obj)
            for mask_name, mask_grp in kwargs.items():
                self.logger.debug(f'Applying "{mask_name}" with parameters {mask_grp}')
                masks = get_masks(self.velocimetry_mask_obj, **mask_grp)
                # apply found masks on velocimetry object
                self.velocimetry_mask_obj.velocimetry.mask(masks, inplace=True)
            self.logger.info("Velocimetry masks applied")
            # set the encoding to a good compression level
            self.velocimetry_mask_obj.velocimetry.set_encoding()
            # store results to file
            if write:
                delayed_obj = self.velocimetry_mask_obj.to_netcdf(self.fn_piv_mask, compute=False)
                with ProgressBar():
                    delayed_obj.compute()
                del delayed_obj
            self.logger.info(f"Velocimetry masked written to {self.fn_piv_mask}.")
        except Exception as e:
            self.logger.error(f"Could not apply masks to velocimetry. Error: {e}")
            raise Exception(f"Could not apply masks to velocimetry. Error: {e}")

    @run_func_hash_io(check=False, configs=["transect"], inputs=["fn_piv_mask"])
    def transect(self, write=False, **kwargs):
        try:
            self.logger.debug("Deriving transects from velocimetry.")
            self.transects = {}
            # keep integrity of original kwargs
            _kwargs = copy.deepcopy(kwargs)
            for transect_name, transect_grp in _kwargs.items():
                self.logger.debug(f'Processing transect "{transect_name}"')
                # check if there are coordinates provided

                if not ("shapefile" in transect_grp or "geojson" in transect_grp):
                    raise click.UsageError(
                        f'Transect with name "{transect_name}" does not have a "shapefile" or '
                        f'"geojson". Please add "shapefile" in the recipe file or provide a '
                        f'shapefile or geojson file with the "--cross" option.'
                    )
                # read geojson or shapefile (as alternative
                if "geojson" in transect_grp:
                    # read directly from geojson
                    coords, crs = cli_utils.read_shape(geojson=transect_grp["geojson"])
                elif "shapefile" in transect_grp:
                    coords, crs = cli_utils.read_shape(fn=transect_grp["shapefile"])
                self.logger.debug(f"Coordinates read for transect {transect_name}")
                # check if coords have z coordinates
                if len(coords[0]) == 2:
                    raise click.UsageError(
                        f'Transect in {os.path.abspath(transect_grp["shapefile"])} only contains x, y, but no '
                        f'z-coordinates.'
                    )
                x, y, z = zip(*coords)
                self.logger.debug(f"Sampling transect {transect_name}")
                # sample the coordinates
                if "get_transect" not in transect_grp:
                    transect_grp["get_transect"] = {}
                if transect_grp["get_transect"] is None:
                    transect_grp["get_transect"] = {}
                self.transects[transect_name] = self.velocimetry_mask_obj.velocimetry.get_transect(
                    x=x, y=y, z=z, crs=crs, **transect_grp["get_transect"]
                )
                if "get_q" in transect_grp:
                    if transect_grp["get_q"] is None:
                        transect_grp["get_q"] = {}
                    # add q
                    self.transects[transect_name] = self.transects[transect_name].transect.get_q(
                        **transect_grp["get_q"]
                    )
                if "get_river_flow" in transect_grp:
                    if "get_q" not in transect_grp:
                        raise click.UsageError(
                            f'"get_river_flow" found in {transect_name} but no '
                            f'"get_q" found, which is a requirement for "get_river_flow"'
                        )
                    if transect_grp["get_river_flow"] is None:
                        transect_grp["get_river_flow"] = {}
                    # add q
                    self.transects[transect_name].transect.get_river_flow(**transect_grp["get_river_flow"])
                if write:
                    # output file
                    fn_transect = os.path.abspath(self.fn_transect_template(transect_name))
                    self.logger.debug(f'Writing transect "{transect_name}" to {fn_transect}')
                    delayed_obj = self.transects[transect_name].to_netcdf(fn_transect, compute=False)
                    with ProgressBar():
                        delayed_obj.compute()
                    self.logger.info(f'Transect "{transect_name}" written to {fn_transect}')
            self.logger.info("Transects derived.")
        except Exception as e:
            self.logger.error(f"Could not derive transects from velocimetry. Error: {e}")
            raise Exception(f"Could not derive transects from velocimetry. Error: {e}")

    @run_func_hash_io(
        check=False,
        configs=["video", "frames", "velocimetry", "transect", "plot"],
        inputs=["fn_video", "fn_piv_mask"],
        outputs=[],
    )
    def plot(self, **plot_recipes):
        try:
            self.logger.debug("Plotting velocimetry.")
            _plot_recipes = copy.deepcopy(plot_recipes)
            for name, plot_params in _plot_recipes.items():
                self.logger.debug(f'Processing plot "{name}"')
                fn_jpg = os.path.join(self.output, self.prefix + name + ".jpg")
                mode = plot_params["mode"]
                ax = None
                # look for inputs
                if "frames" in plot_params:
                    if "frame_number" in plot_params:
                        n = plot_params["frame_number"]
                    else:
                        n = 0
                    opts = plot_params["frames"] if plot_params["frames"] is not None else {}
                    f = self.video_obj.get_frames(method="rgb")
                    if mode != "camera":
                        f = f[n : n + 1].frames.project(method=self.proj_method)[0]
                    else:
                        f = f[n]
                    p = f.frames.plot(ax=ax, mode=mode, **opts)
                    # continue with axes of p
                    ax = p.axes
                if "velocimetry" in plot_params:
                    opts = plot_params["velocimetry"]
                    opts = vmin_vmax_to_norm(opts)
                    # select the time reducer. If not defined, choose mean
                    reducer = plot_params["reducer"] if "reducer" in plot_params else "mean"
                    reducer_params = plot_params["reducer_params"] if "reducer_params" in plot_params else {}
                    # reduce the velocimetry over time for plotting purposes
                    velocimetry_reduced = getattr(self.velocimetry_mask_obj, reducer)(
                        dim="time", keep_attrs=True, **reducer_params
                    )
                    p = velocimetry_reduced.velocimetry.plot(ax=ax, mode=mode, **opts)
                    ax = p.axes
                    del velocimetry_reduced
                if "transect" in plot_params:
                    for transect_name, opts in plot_params["transect"].items():
                        opts = vmin_vmax_to_norm(opts)
                        # read file
                        fn_transect = self.fn_transect_template(transect_name)
                        ds_trans = xr.open_dataset(fn_transect)
                        # default quantile is 2 (50%), otherwise choose from list
                        quantile = 2 if ("quantile") not in opts else opts["quantile"]
                        ds_trans_q = ds_trans.isel(quantile=quantile)
                        # add to plot
                        p = ds_trans_q.transect.plot(ax=ax, mode=mode, **opts)
                        ax = p.axes
                        # done with transect, remove from memory
                        ds_trans.close()
                        del ds_trans
                # if mode == "camera":
                #     ax.axis("equal")
                write_pars = plot_params["write_pars"] if "write_pars" in plot_params else {}
                self.logger.debug(f'Writing plot "{name}" to {fn_jpg}')
                ax.figure.savefig(fn_jpg, **write_pars)
                self.logger.info(f'Plot "{name}" written to {fn_jpg}')
            self.logger.info("Plot procedure done.")
        except Exception as e:
            self.logger.error(f"Could not plot velocimetry. Error: {e}")
            raise Exception(f"Could not plot velocimetry. Error: {e}")


def velocity_flow(**kwargs):
    """Execute the entire velocimetry processor."""
    processor = VelocityFlowProcessor(**kwargs)
    # process video following the settings
    processor.process()
    # flush the processor itself once done
    del processor


def velocity_flow_subprocess(
    videofile,
    recipe,
    cameraconfig,
    output,
    prefix=None,
    h_a: float = None,
    cross: str = None,
    update: bool = False,
    concurrency=True,
    logger=logging,
):
    """Write the requirements to temporary files and run velocimetry from a separate CLI instance.

    This is used for automated running of velocimetry on videos from other applications.

    Parameters
    ----------
    recipe : dict
        YAML recipe, parsed from CLI
    videofile : str
        path to video
    cameraconfig : dict
        camera config as dict (not yet loaded as CamerConfig object)
    prefix : str
        prefix of produced output files
    h_a : float, optional
        Current water level in meters (default None)
    cross : dict, optional
        Cross-section profile as geojson dict, to use for optical water level detection if provided and h_a is None,
    output : str
        path to output file
    update : bool, optional
        if set, only update components with changed inputs and configurations
    concurrency : bool, optional
        if set to False, then dask will only run synchronous preventing overuse of memory. This will be slower
    logger : logging
        reference to logger instance

    Returns
    -------
    subprocess.CompletedProcess
        Process status

    """
    # store recipe in file
    logger.info(f"Launching separate pyorc instance for videofile {videofile}")
    if not (os.path.isdir(output)):
        os.makedirs(output)
    fn_recipe = os.path.join(output, "recipe.yml")
    fn_cam_config = os.path.join(output, "camera_config.json")
    with open(fn_recipe, "w") as f:
        yaml.dump(recipe, f, default_flow_style=False, sort_keys=False)
    CameraConfig(**cameraconfig).to_file(fn_cam_config)
    cmd = ["pyorc", "velocimetry", "-V", videofile, "-c", fn_cam_config, "-r", fn_recipe]
    # add components where needed
    if h_a is not None:
        cmd.append("-h")
        cmd.append(str(h_a))
    if h_a is None and cross is not None:
        cross_fn = os.path.join(output, "cross.geojson")
        # write cross-section to file
        with open(cross_fn, "w") as f:
            json.dump(cross, f, indent=4)
        cmd.append("--cross")
        cmd.append(cross_fn)
    if not concurrency:
        cmd.append("--lowmem")
    if update:
        cmd.append("-u")
    if prefix:
        cmd.append("-p")
        cmd.append(prefix)
    cmd_suffix = ["-u", "-vvv", output]
    cmd = cmd + cmd_suffix
    # call subprocess
    result = subprocess.run(cmd)
    return result



"""Plot methods for pyorc."""

import copy
import functools

import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import numpy as np
import warnings

from matplotlib import colors, patheffects

from pyorc import helpers

path_effects = [
    patheffects.Stroke(linewidth=3, foreground="w"),
    patheffects.Normal(),
]

LINE_COLOR = "#385895"


def _base_plot(plot_func):
    commondoc = """

    Parameters
    ----------
    mode : str, optional
        perspective mode to use for plotting. Can be "local" (default), "geographical", or "camera". For
        "geographical" a velocimetry result that contains "lon" and "lat" coordinates must be provided
        (i.e. produced with known CRS for control points).
    x : str, optional
        Coordinate for *x* axis. If ``None``, use ``darray.dims[1]``.
    y : str, optional
        Coordinate for *y* axis. If ``None``, use ``darray.dims[0]``.
    ax : plt.axes, optional
        If None (default), use the current axes. Not applicable when using facets.
    **kwargs : additional keyword arguments to wrapped Matplotlib function.

    Returns
    -------
    artist : matplotlib mappable
        The same type of primitive artist that the wrapped Matplotlib
        function returns.

    """
    # This function is largely based on xarray.Dataset function _dsplot
    # Build on the original docstring
    plot_func.__doc__ = f"{plot_func.__doc__}{commondoc}"

    # apply wrapper to allow for partial update of the function, with updated docstring
    @functools.wraps(plot_func)
    def get_plot_method(
        ref,
        mode="local",
        ax=None,
        add_colorbar=False,
        add_cross_section=True,
        add_text=False,
        text_prefix="",
        text_suffix="",
        kwargs_line=None,
        **kwargs,
    ):
        """Retrieve plot method with all required inputs.

        Parameters
        ----------
        ref : xr.Dataset
            velocimetry or transect object
        mode : str, optional
            perspective mode to use for plotting. Can be "local", "geographical", or "camera". For
            "geographical" a velocimetry result that contains "lon" and "lat" coordinates must be provided
            (i.e. produced with known CRS for control points).
        ax : plt.axes object, optional
            If None (default), use the current axes. Not applicable when using facets.
        add_colorbar : boolean, optional
            if True, a colorbar is added to axes (default: False)
        add_cross_section : boolean, optional
            if True, and a transect is plotted, the transect coordinates are plotted (default: True)
        add_text : boolean, optional
            if True, add a text label in the axes displaying information about the video's transect
        text_prefix : str, optional
            string to add in front of standard text on transect plot. Only used if `add_text=True`
        text_suffix : str, optional
            String to add after standard text on transect plot. Only used if `add_text=True`
        kwargs_line : dict, optional
            additional keyword arguments passed to matplotlib.pyplot.plot for plotting cross-section.
            (Default value = {})
        **kwargs : dict
            additional keyword arguments to wrapped Matplotlib function.

        Returns
        -------
        p : matplotlib mappable
            mappable of wrapped matplotlib function

        """
        if not kwargs_line:
            kwargs_line = {}
        # in case persistent transform is in kwargs_line, remove this
        if mode == "geographical":
            try:
                pass
            except ImportError:
                raise ImportError("Cartopy not found, please install with 'mamba install -c conda-forge cartopy")
        if "transform" in kwargs_line:
            del kwargs_line["transform"]
        ax = _prepare_axes(ax=ax, mode=mode)
        # update ax

        if len(ref._obj["v_x"].shape) > 2:
            raise OverflowError(
                f'Dataset\'s variables should only contain 2 dimensions, this dataset '
                f'contains {len(ref._obj["v_x"].shape)} dimensions. Apply a reducer or select a time step. '
                f'Reducing can be done e.g. with ds.mean(dim="time", keep_attrs=True) or slicing with ds.isel(time=0)'
            )
        if "time" in ref._obj:
            if isinstance(ref._obj["time"].values.tolist(), list):
                raise OverflowError(
                    'Dimension "time" exists in dataset and has multiple values. Reduce this by applying a reducer or '
                    "selecting a time step."
                )
        if "quantile" in ref._obj:
            if isinstance(ref._obj["quantile"].values.tolist(), list):
                raise OverflowError(
                    'Dimension "quantile" exists in dataset and contains multiple values. Reduce this by selecting one'
                    "quantile, e.g. using ds.isel(quantile=2) or ds.sel(quantile=0.5)"
                )
        if plot_func.__name__ == "streamplot":
            if mode != "local":
                raise NotImplementedError(f"Streamplot only works in local mode, not in {mode} mode.")

        # check if dataset is a transect or not
        is_transect = True if "points" in ref._obj.dims else False
        kwargs = set_default_kwargs(kwargs, method=plot_func.__name__, mode=mode)
        assert mode in ["local", "geographical", "camera"], 'Mode must be "local", "geographical" or "camera"'
        if mode == "local":
            x = ref._obj["x"].values
            y = ref._obj["y"].values
            u, v, s = ref.get_uv_local()
            if not (is_transect) and plot_func.__name__ == "scatter":
                x, y = np.meshgrid(x, y)

            if plot_func.__name__ == "streamplot":
                # flipping of variables is needed
                y = np.flipud(y)
                u = np.flipud(u)
                v = np.flipud(v)
                s = np.flipud(s)

        elif mode == "geographical":
            # import some additional packages
            import cartopy.crs as ccrs

            # add transform for GeoAxes
            kwargs["transform"] = ccrs.PlateCarree()
            kwargs_line["transform"] = ccrs.PlateCarree()

            x = ref._obj["lon"].values
            y = ref._obj["lat"].values
            u, v, s = ref.get_uv_geographical()
        else:
            # mode is camera
            x = ref._obj["xp"].values
            y = ref._obj["yp"].values
            u, v, s = ref.get_uv_camera()
        if plot_func.__name__ in ["quiver", "streamplot"]:
            primitive = plot_func("", x, y, u, v, s, ax, **kwargs)
        else:
            primitive = plot_func("", x, y, s, ax, **kwargs)
        if add_colorbar:
            cbar(ax, primitive)
        if mode == "local":
            ax.set_aspect("equal")
        if is_transect:
            if add_cross_section:
                if mode == "camera":
                    # # lens position is needed, so check this
                    ref._obj.transect.cross_section.plot(
                        h=ref._obj.transect.h_a,
                        ax=ax,
                        camera=True,
                        swap_y_coords=True,
                        wetted=True,
                        planar=False,
                        bottom=False,
                    )
                    # check if water level is above the lowest level
                    check_low = (
                        ref._obj.transect.camera_config.h_to_z(ref._obj.transect.h_a)
                        > ref._obj.transect.cross_section.z.min()
                    )
                    check_high = (
                        ref._obj.transect.camera_config.h_to_z(ref._obj.transect.h_a)
                        < ref._obj.transect.cross_section.z.max()
                    )
                    if check_low and check_high:
                        try:
                            ref._obj.transect.cross_section.plot_water_level(
                                h=ref._obj.transect.h_a,
                                length=2.0,
                                linewidth=3.0,
                                ax=ax,
                                camera=True,
                                swap_y_coords=True,
                                color="r",
                                label="water level",
                            )
                        except:
                            warnings.warn(
                                "Not able to find a unique location for plotting of water level",
                                stacklevel=2
                            )

                    # draw some depth lines for better visual interpretation.
                    depth_lines = ref._obj.transect.get_depth_perspective(h=ref._obj.transect.h_a)
                    for l in depth_lines:
                        line = np.array(l)
                        ax.plot(
                            line[:, 0],
                            line[:, 1],
                            color="w",
                            alpha=0.5,
                            linewidth=2.0,
                            zorder=1,
                            # path_effects=path_effects
                        )
                else:
                    ax.plot(x, y, LINE_COLOR, path_effects=path_effects, alpha=0.7, **kwargs_line)
                if add_text:
                    plot_text(ax, ref._obj, text_prefix, text_suffix)

        if mode == "geographical" and not (is_transect):
            ax.set_extent(
                [x.min() - 0.0001, x.max() + 0.0001, y.min() - 0.0001, y.max() + 0.0001], crs=ccrs.PlateCarree()
            )
        return primitive

    return get_plot_method


def _frames_plot(ref, ax=None, mode="local", **kwargs):
    """Create QuadMesh plot from a RGB or grayscale frame on a new or existing (if ax is not None) axes.

    Wraps :py:func:`matplotlib:matplotlib.collections.QuadMesh`.

    Parameters
    ----------
    ref : plot reference
        This parameter should not be specified, ignore.
    ax : plt.axes, optional
        If None (default), use the current axes. Not applicable when using facets.
    mode : str, optional
        perspective mode to use for plotting. Can be "local" (default), "geographical", or "camera".
        For "geographical" a frames set from Frames.project should be used that contains "lon" and "lat" coordinates.
        For "camera", a non-projected frames set should be used.
    **kwargs : dict
        additional keyword arguments to wrapped Matplotlib function.

    Returns
    -------
    p : matplotlib.collections.QuadMesh

    """  # noqa: D417
    # prepare axes
    if mode == "geographical":
        try:
            pass
        except ImportError:
            raise ImportError("Cartopy not found, please install with 'mamba install -c conda-forge cartopy")
    if "time" in ref._obj.coords:
        if ref._obj.time.size > 1:
            raise AttributeError(
                f'Object contains dimension "time" with length {len(ref._obj.time)}. Reduce dataset by selecting'
                "one time step or taking a median, mean or other statistic."
            )
    if mode == "camera":
        rotation = ref.camera_config.rotation
    else:
        rotation = None
    ax = _prepare_axes(ax=ax, mode=mode, rotation=rotation)
    if mode == "local":
        x = "x"
        y = "y"
    elif mode == "geographical":
        # import some additional packages
        import cartopy.crs as ccrs

        # add transform for GeoAxes
        kwargs["transform"] = ccrs.PlateCarree()
        x = "lon"
        y = "lat"
    else:
        # mode is camera
        x = "xp"
        y = "yp"
    assert all(v in ref._obj.coords for v in [x, y]), f'required coordinates "{x}" and/or "{y}" are not available'
    if x == "x":
        # use a simple imshow, much faster
        dx = abs(float(ref._obj[x][1] - ref._obj[x][0]))  # x grid cell size
        dy = abs(float(ref._obj[y][1] - ref._obj[y][0]))  # y grid cell size
        # Calculate extent with half grid cell expansion
        extent = [
            ref._obj[x].min().item() - 0.5 * dx,
            ref._obj[x].max().item() + 0.5 * dx,
            ref._obj[y].min().item() - 0.5 * dy,
            ref._obj[y].max().item() + 0.5 * dy,
        ]

    if x != "x":
        primitive = ax.pcolormesh(ref._obj[x], ref._obj[y], ref._obj, **kwargs)
    else:
        primitive = ax.imshow(ref._obj, origin="upper", extent=extent, aspect="auto", **kwargs)
    # fix axis limits to min and max of extent of frames
    if mode == "geographical":
        ax.set_extent(
            [
                ref._obj[x].min() - 0.0001,
                ref._obj[x].max() + 0.0001,
                ref._obj[y].min() - 0.0001,
                ref._obj[y].max() + 0.0001,
            ],
            crs=ccrs.PlateCarree(),
        )
    else:
        ax.set_xlim([ref._obj[x].min(), ref._obj[x].max()])
        ax.set_ylim([ref._obj[y].min(), ref._obj[y].max()])
    return primitive


class _Transect_PlotMethods:
    """Enable use of `ds.velocimetry.plot` functions as attributes on a Dataset containing velocimetry results.

    For example, `ds.velocimetry.plot.pcolormesh`. When called without a subfunction, quiver will be used.
    """

    def __init__(self, transect):
        # make the original dataset also available on the plotting object
        self.transect = transect
        self._obj = transect._obj
        # Add to class _PlotMethods

    def __call__(self, method="quiver", **kwargs):
        """call.

        Parameters
        ----------
        method : str, optional
            plot method to use (default: "quiver")
        **kwargs : dict
            additional keyword arguments to wrapped Matplotlib function.

        Returns
        -------
        p : matplotlib mappable
            mappable of wrapped matplotlib function

        """
        return getattr(self, method)(**kwargs)

    def get_uv_camera(self, dt=0.1):
        """Get x-directional (u), y-directional (v) and scalar velocity in camera projection from transect dataset.

        Parameters
        ----------
        dt : float, optional
            time step [s] used to transpose velocities over a short distance for projection (default: 0.1)

        Returns
        -------
        u : np.ndarray
            x-directional velocity
        v : np.ndarray
            y-directional velocity
        s : np.ndarray
            scalar velocity

        """
        v_eff = "v_eff"
        v_dir = "v_dir"
        # retrieve the backward transformation array
        transect = self._obj.transect
        camera_config = transect.camera_config
        x, y = self._obj.x, self._obj.y

        _u = self._obj[v_eff] * np.sin(self._obj[v_dir])
        _v = self._obj[v_eff] * np.cos(self._obj[v_dir])
        s = np.abs(self._obj[v_eff].values)
        # x_moved, y_moved = x + _u * dt, y + _v * dt
        x_moved, y_moved = x + _u, y + _v
        # transform to real-world
        cols_moved, rows_moved = x_moved / camera_config.resolution, y_moved / camera_config.resolution
        rows_moved = camera_config.shape[0] - rows_moved
        xs_moved, ys_moved = helpers.get_xs_ys(cols_moved, rows_moved, camera_config.transform)
        cols, rows = x / camera_config.resolution, y / camera_config.resolution
        rows = camera_config.shape[0] - rows
        xs, ys = helpers.get_xs_ys(cols, rows, camera_config.transform)
        xp_moved, yp_moved = list(
            zip(
                *camera_config.project_points(
                    list(map(list, zip(xs_moved, ys_moved, np.ones(x.shape) * camera_config.h_to_z(transect.h_a)))),
                    swap_y_coords=True,
                )
            )
        )
        xp, yp = list(
            zip(
                *camera_config.project_points(
                    list(map(list, zip(xs, ys, np.ones(x.shape) * camera_config.h_to_z(transect.h_a)))),
                    swap_y_coords=True,
                )
            )
        )
        xp, yp, xp_moved, yp_moved = np.array(xp), np.array(yp), np.array(xp_moved), np.array(yp_moved)
        # remove vectors that have nan on moved pixels
        xp_moved[np.isnan(x_moved)] = np.nan
        yp_moved[np.isnan(y_moved)] = np.nan

        self._obj["xp"][:] = xp  # [:].flatten()
        self._obj["yp"][:] = yp  # [:].flatten()
        u, v = xp_moved - self._obj["xp"], yp_moved - self._obj["yp"]
        return u, v, s

    def get_uv_geographical(self):
        """Get x-directional (u), y-directional (v) and scalar velocity in geo projection from transect dataset.

        Returns
        -------
        u : np.ndarray
            x-directional velocity
        v : np.ndarray
            y-directional velocity
        s : np.ndarray
            scalar velocity

        """
        v_eff = "v_eff"
        v_dir = "v_dir"
        u = self._obj[v_eff] * np.sin(self._obj[v_dir])
        v = self._obj[v_eff] * np.cos(self._obj[v_dir])
        s = self._obj[v_eff]
        aff = self.transect.camera_config.transform
        theta = np.arctan2(aff.d, aff.a)
        # rotate velocity vectors along angle theta to match the requested projection. this only changes values
        # in case of camera projections
        u, v = helpers.rotate_u_v(u, v, theta)
        return u, v, s

    def get_uv_local(self):
        """Get x-directional (u), y-directional (v) and scalar velocity in local projection from transect dataset.

        Returns
        -------
        u : np.ndarray
            x-directional velocity
        v : np.ndarray
            y-directional velocity
        s : np.ndarray
            scalar velocity

        """
        v_eff = "v_eff"
        v_dir = "v_dir"
        u = self._obj[v_eff] * np.sin(self._obj[v_dir])
        v = self._obj[v_eff] * np.cos(self._obj[v_dir])
        s = self._obj[v_eff]
        return u, v, s


class _Velocimetry_PlotMethods:
    """Enable use of `ds.velocimetry.plot` functions as attributes on a Dataset containing velocimetry results.

    For example, `ds.velocimetry.plot.pcolormesh`. When called without a subfunction, quiver will be used
    """

    def __init__(self, velocimetry):
        # make the original dataset also available on the plotting object
        self.velocimetry = velocimetry
        self._obj = velocimetry._obj
        # Add to class _PlotMethods

    def __call__(self, method="quiver", **kwargs):
        """call.

        Parameters
        ----------
        method : str, optional
            plot method to use "quiver", "scatter", "streamplot" or "pcolormesh" (default: "quiver")
        **kwargs : dict
            additional keyword arguments to wrapped Matplotlib function.

        Returns
        -------
        p : matplotlib mappable
            mappable of wrapped matplotlib function

        """
        return getattr(self, method)(**kwargs)

    def get_uv_geographical(self):
        """Get x-directional (u), y-directional (v) and scalar velocity in camera projection from velocimetry dataset.

        Returns
        -------
        u : np.ndarray
            x-directional velocity
        v : np.ndarray
            y-directional velocity
        s : np.ndarray
            scalar velocity

        """
        # select lon and lat variables as coordinates
        velocimetry = self._obj.velocimetry
        v_x = self._obj["v_x"].values
        v_y = self._obj["v_y"].values
        u = v_x / (2 * 1e5)  # 1e5 is aobut the amount of meters per degree
        v = -v_y / (2 * 1e5)
        s = (v_x**2 + v_y**2) ** 0.5
        aff = velocimetry.camera_config.transform
        theta = np.arctan2(aff.d, aff.a)
        # rotate velocity vectors along angle theta to match the requested projection. this only changes values
        # in case of camera projections
        u, v = helpers.rotate_u_v(u, v, theta)
        return u, v, s

    def get_uv_local(self):
        """Get x-directional (u), y-directional (v) and scalar velocity in local projection from velocimetry dataset.

        Returns
        -------
        u : np.ndarray
            x-directional velocity
        v : np.ndarray
            y-directional velocity
        s : np.ndarray
            scalar velocity

        """
        # u and v should be scaled to a nice proportionality for local plots. This is typically about 2x smaller
        # than the m/s values
        v_x = self._obj["v_x"].values
        v_y = self._obj["v_y"].values

        u = v_x / 2
        v = -v_y / 2
        s = (v_x**2 + v_y**2) ** 0.5
        return u, v, s

    def get_uv_camera(self, dt=0.1):
        """Get x-directional (u), y-directional (v) and scalar velocity in camera projection from velocimetry dataset.

        Parameters
        ----------
        dt : float, optional
            time step [s] used to transpose velocities over a short distance for projection (default: 0.1)

        Returns
        -------
        u : np.ndarray
            x-directional velocity
        v : np.ndarray
            y-directional velocity
        s : np.ndarray
            scalar velocity

        """
        # retrieve the backward transformation array from x, y to persective row column
        velocimetry = self._obj.velocimetry
        camera_config = velocimetry.camera_config
        # get the shape of the original frames
        shape_y, shape_x = velocimetry.camera_shape
        xi, yi = np.meshgrid(self._obj.x, self._obj.y)
        # flip the y-coordinates to match the row order used by opencv
        yi = np.flipud(yi)

        # follow the velocity vector over a short distance (dt*velocity)
        # x_moved, y_moved = xi + self._obj["v_x"] * dt, yi + self._obj["v_y"] * dt
        x_moved, y_moved = xi + self._obj["v_x"] / 2, yi + self._obj["v_y"] / 2
        # transform to real-world
        cols_moved, rows_moved = x_moved / camera_config.resolution, y_moved / camera_config.resolution
        xs_moved, ys_moved = helpers.get_xs_ys(cols_moved, rows_moved, camera_config.transform)
        cols, rows = xi / camera_config.resolution, yi / camera_config.resolution
        xs, ys = helpers.get_xs_ys(cols, rows, camera_config.transform)
        # project the found displacement points to camera projection
        xp_moved, yp_moved = camera_config.project_grid(
            xs_moved, ys_moved, np.ones(xi.shape) * camera_config.h_to_z(velocimetry.h_a), swap_y_coords=True
        )
        xp, yp = camera_config.project_grid(
            xs, ys, np.ones(xi.shape) * camera_config.h_to_z(velocimetry.h_a), swap_y_coords=True
        )

        # missing values end up at the top-left, replace these with nan
        yp_moved[yp_moved == shape_y] = np.nan  # ds["yp"].values[yp_moved == shape_y]
        xp_moved[xp_moved == 0] = np.nan  # ds["xp"].values[xp_moved == 0]

        # estimate the projected velocity vector
        u, v = xp_moved - xp, yp_moved - yp
        self._obj["xp"][:] = xp[:]
        self._obj["yp"][:] = yp[:]
        s = ((self._obj["v_x"] ** 2 + self._obj["v_y"] ** 2) ** 0.5).values
        return u, v, s


def set_default_kwargs(kwargs, method="quiver", mode="local"):
    """Set color mapping default kwargs if no vmin and/or vmax is supplied."""
    if mode == "local":
        # width scale is in cm
        width_scale = 0.02
    elif mode == "geographical":
        # widths are in degrees
        width_scale = 0.00000025
    elif mode == "camera":
        # widths in pixels
        width_scale = 3.0
    else:
        raise ValueError("mode must be one of 'local', 'geographical' or 'camera'")
    if "cmap" not in kwargs:
        kwargs["cmap"] = "rainbow"  # the famous rainbow colormap!
    if "vmin" not in kwargs and "vmax" not in kwargs and "norm" not in kwargs:
        # set a normalization array
        norm = [0, 0.05, 0.1, 0.2, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0]
        kwargs["norm"] = colors.BoundaryNorm(norm, ncolors=256, extend="max")
    if method == "quiver":
        if "scale" not in kwargs:
            kwargs["scale"] = 1  # larger quiver arrows
        if "units" not in kwargs:
            kwargs["units"] = "xy"
        # for width, it will matter a lot what mode is used, width should be a dimensionless number
        if "width" not in kwargs:
            kwargs["width"] = width_scale
        else:
            kwargs["width"] *= width_scale  # multiply

    return kwargs


@_base_plot
def quiver(_, x, y, u, v, s=None, ax=None, **kwargs):
    """Create quiver plot from velocimetry results on new or existing axes.

    Note that the `width` parameter is always in a unitless scale, with `1` providing a nice-looking default value
    for all plot modes. The default value is usually very nice. If you do want to change:

    The `scale` parameter defaults to 1, providing a nice looking arrow length. With a smaller (larger) value,
    quivers will become longer (shorter).

    The `width` parameter is in a unitless scale, with `1` providing a nice-looking default value.

    Wraps :py:func:`matplotlib:matplotlib.pyplot.quiver`.
    """
    if "color" in kwargs:
        # replace scalar colors by one single color
        primitive = ax.quiver(x, y, u, v, **kwargs)
    else:
        primitive = ax.quiver(x, y, u, v, s, **kwargs)
    #
    return primitive


@_base_plot
def scatter(_, x, y, c=None, ax=None, **kwargs):
    """Create scatter plot of velocimetry or transect results on new or existing axes.

    Wraps :py:func:`matplotlib:matplotlib.pyplot.scatter`.
    """
    primitive = ax.scatter(x, y, c=c, **kwargs)
    return primitive


@_base_plot
def streamplot(_, x, y, u, v, s=None, ax=None, linewidth_scale=None, **kwargs):
    """Create streamplot of velocimetry results on new or existing axes.

    Wraps :py:func:`matplotlib:matplotlib.pyplot.streamplot`. Additional input arguments:
    """
    if linewidth_scale is not None:
        kwargs["linewidth"] = s * linewidth_scale
    if "color" in kwargs:
        primitive = ax.streamplot(x, y, u, v, **kwargs)
    else:
        primitive = ax.streamplot(x, y, u, v, color=s, **kwargs)
    return primitive.lines


@_base_plot
def pcolormesh(_, x, y, s=None, ax=None, **kwargs):
    """Create pcolormesh plot from velocimetry results on new or existing axes.

    Wraps :py:func:`matplotlib:matplotlib.pyplot.pcolormesh`.
    """
    primitive = ax.pcolormesh(x, y, s, **kwargs)
    return primitive


def cbar(ax, p, size=12, **kwargs):
    """Add colorbar to existing axes.

    In case camera mode is used, the colorbar will get a bespoke layout and will be placed inside the axes object.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object where the colorbar will be added.
    p : matplotlib mappable
        used to define colorbar
    size : float, optional
        fontsize, used for colorbar title
    **kwargs :
        dict, additional settings passed to plt.colorbar

    Returns
    -------
    cbar : matplotlib colorbar
        handle to colorbar

    """
    label_format = "{:,.2f}"
    path_effects = [
        patheffects.Stroke(linewidth=2, foreground="w"),
        patheffects.Normal(),
    ]
    cax = ax.inset_axes([0.05, 0.05, 0.02, 0.25])
    cb = ax.figure.colorbar(p, cax=cax, **kwargs)
    ticks_loc = cb.get_ticks().tolist()
    cb.set_ticks(mticker.FixedLocator(ticks_loc))
    cb.set_ticklabels([label_format.format(x) for x in ticks_loc], path_effects=path_effects, fontsize=size)
    cb.set_label(label="velocity [m/s]", size=size, path_effects=path_effects)
    return cb


def plot_text(ax, ds, prefix, suffix):
    """Add text with info on transect to plot in standardized manner.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object where the text will be plotted.
    ds : xarray.Dataset
        The dataset containing transect and river flow information.
    prefix : str
        The string that will appear before the main content of the text.
    suffix : str
        The string that will appear after the main content of the text.

    """
    if "q" not in ds:
        return
    _ds = copy.deepcopy(ds)
    xloc = 0.95
    yloc = 0.95
    _ds.transect.get_river_flow(q_name="q")
    Q = np.abs(_ds.river_flow)
    string = prefix
    string += "Water level: {:1.2f} m\nDischarge: {:1.2f} m3/s".format(_ds.transect.h_a, Q.values)
    if "q_nofill" in ds:
        _ds.transect.get_river_flow(q_name="q_nofill")
        Q_nofill = np.abs(_ds.river_flow)
        perc_measured = Q_nofill / Q * 100  # fraction that is truly measured compared to total
        string += " ({:1.0f}% measured)".format(perc_measured.values)
    # reset river flow if necessary
    string += suffix
    ax.text(
        xloc,
        yloc,
        string,
        size=24,
        horizontalalignment="right",
        verticalalignment="top",
        path_effects=path_effects,
        transform=ax.transAxes,
    )


def _prepare_axes(ax=None, mode="local", rotation=None):
    """Prepare the axes, needed to plot results, called from `pyorc.PIV.plot`.

    Parameters
    ----------
    ref : Frames, Transect or Velocimetry
        object to derive plot from
    ax : plt.axes, optional
        if not provided (default), a new axes is prepared (default: None)
    mode : str, optional
        mode to plot, can be "local" (default), "geographical" or "camera".
    rotation : Int[0, 90, 180, 270], optional
        Rotation of image (if vertically oriented, axis sizes are also rotated).


    Returns
    -------
    ax : plt.axes

    """
    if ax is not None:
        if mode == "geographical":
            # ensure that the axes is a geoaxes
            from cartopy.mpl.geoaxes import GeoAxesSubplot

            assert isinstance(
                ax, GeoAxesSubplot
            ), "For mode=geographical, the provided axes must be a cartopy GeoAxesSubplot"
        return ax

    # make a screen filling figure with black edges and faces
    if rotation in [90, 270]:
        f = plt.figure(figsize=(9, 16), frameon=False, facecolor="k")
        f.set_size_inches(9, 16, True)
    else:
        f = plt.figure(figsize=(16, 9), frameon=False, facecolor="k")
        f.set_size_inches(16, 9, True)

    f.patch.set_facecolor("k")
    f.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
    if mode == "geographical":
        import cartopy.crs as ccrs

        ax = f.add_subplot(111, projection=ccrs.PlateCarree())
    else:
        ax = plt.subplot(111)
    return ax


# add all required plot methods below
_Velocimetry_PlotMethods.quiver = quiver
_Velocimetry_PlotMethods.pcolormesh = pcolormesh
_Velocimetry_PlotMethods.scatter = scatter
_Velocimetry_PlotMethods.streamplot = streamplot
_Transect_PlotMethods.quiver = quiver
_Transect_PlotMethods.scatter = scatter



"""Main command line routines pyorc."""

import os
import sys
from typing import List, Optional, Union

import click
from typeguard import typechecked

from pyorc import __version__, service
from pyorc.cli import cli_utils, log


def print_info(ctx, param, value):
    """Display pyorc information."""
    if not value:
        return {}
    click.echo(f"PyOpenRiverCam version {__version__}, Copyright Localdevices, Rainbow Sensing\n")
    ctx.exit()


def print_license(ctx, param, value):
    """Display pyorc license."""
    if not value:
        return {}
    click.echo("GNU Affero General Public License v3 (AGPLv3). See https://www.gnu.org/licenses/agpl-3.0.en.html\n")
    ctx.exit()


verbose_opt = click.option("--verbose", "-v", count=True, help="Increase verbosity.")
video_opt = click.option(
    "-V",
    "--videofile",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False, file_okay=True),
    help="video file with required objective and resolution and control points in view",
    callback=cli_utils.validate_file,
    required=True,
)


@click.group(context_settings={"max_content_width": 120})
@click.version_option(__version__, message="PyOpenRiverCam version: %(version)s")
@click.option(
    "--info",
    default=False,
    is_flag=True,
    is_eager=True,
    help="Print information and version of PyOpenRiverCam",
    callback=print_info,
)
@click.option(
    "--license",
    default=False,
    is_flag=True,
    is_eager=True,
    help="Print license information for PyOpenRiverCam",
    callback=print_license,
)
@click.pass_context
def cli(ctx, info, license):
    """Command line interface for pyOpenRiverCam."""
    if ctx.obj is None:
        ctx.obj = {}


# CAMERA CONFIG
@cli.command(short_help="Prepare Camera Configuration file")
@click.argument(
    "OUTPUT",
    type=click.Path(resolve_path=True, dir_okay=False, file_okay=True),
    required=True,
)
@video_opt
@click.option(
    "--crs",
    type=str,
    callback=cli_utils.parse_str_num,
    help="Coordinate reference system to be used for camera configuration",
)
@click.option(
    "-f", "--frame-sample", type=int, default=0, help="Frame number to use for camera configuration background"
)
@click.option(
    "--src", type=str, callback=cli_utils.parse_src, help="Source control points as list of [column, row] pairs."
)
@click.option(
    "--dst",
    type=str,
    callback=cli_utils.parse_dst,
    help="Destination control points as list of 2 or 4 [x, y] pairs, or at least 6 [x, y, z]. If --crs_gcps is "
    "provided, --dst is assumed to be in this CRS.",
)
@click.option("--z_0", type=float, help="Water level [m] +CRS (e.g. geoid or ellipsoid of GPS)")
@click.option("--h_ref", type=float, help="Water level [m] +local datum (e.g. staff or pressure gauge)")
@click.option(
    "--crs_gcps",
    type=str,
    callback=cli_utils.parse_str_num,
    help="Coordinate reference system in which destination GCP points (--dst) are measured",
)
@click.option("--resolution", type=float, help="Target resolution [m] for ortho-projection.")
@click.option("--focal_length", type=float, help="Focal length [pix] of lens.")
@click.option(
    "--k1",
    type=float,
    help="First lens radial distortion coefficient k1 [-]. See also https://docs.opencv.org/2.4/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html",
)
@click.option(
    "--k2",
    type=float,
    help="Second lens radial distortion coefficient k2 [-]. See also https://docs.opencv.org/2.4/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html",
)
@click.option(
    "--window_size", type=int, help="Target window size [px] for interrogation window for Particle Image Velocimetry"
)
@click.option(
    "--shapefile",
    type=click.Path(resolve_path=True, dir_okay=False, file_okay=True),
    help="Shapefile or other GDAL compatible vector file containing dst GCP points [x, y] or [x, y, z] in its geometry",
    callback=cli_utils.validate_file,
)
@click.option(
    "--lens_position",
    type=str,
    help="Lens position as [x, y, z]. If --crs_gcps is provided, --lens_position is assumed to be in this CRS.",
    callback=cli_utils.parse_json,
)
@click.option(
    "--corners",
    type=str,
    callback=cli_utils.parse_corners,
    help="Video ojective corner points as list of 4 [column, row] points",
)
@click.option(
    "--stabilize",
    "-s",
    is_flag=True,
    default=False,
    help="Stabilize the videos using this camera configuration (you can provide a stable area in an interactive view).",
)
@click.option(
    "--rotation",
    type=int,
    required=False,
    callback=cli_utils.validate_rotation,
    help="Provide a rotation of either 90, 180 or 170 degrees if needed to correctly rotate the video",
)
@verbose_opt
@click.pass_context
@typechecked
def camera_config(
    ctx,
    output: str,
    videofile: str,
    crs: Optional[Union[str, int]],
    frame_sample: Optional[int],
    src: Optional[List[List[float]]],
    dst: Optional[List[List[float]]],
    z_0: Optional[float],
    h_ref: Optional[float],
    crs_gcps: Optional[Union[str, int]],
    focal_length: Optional[float],
    k1: Optional[float],
    k2: Optional[float],
    resolution: Optional[float],
    window_size: Optional[int],
    lens_position: Optional[List[float]],
    shapefile: Optional[str],
    corners: Optional[List[List]],
    stabilize: Optional[bool],
    rotation: Optional[int],
    verbose: int,
):
    """CLI subcommand for camera configuration."""
    log_level = max(10, 20 - 10 * verbose)
    logger = log.setuplog("cameraconfig", os.path.abspath("pyorc.log"), append=False, log_level=log_level)
    logger.info(f"Preparing your cameraconfig file in {output}")
    logger.info(f"Found video file  {videofile}")

    if src is not None:
        logger.info("Source points found and validated")
    if dst is not None:
        logger.info("Destination points found and validated")
    if z_0 is None:
        z_0: float = click.prompt("--z_0 not provided, please enter a number, or Enter for default", default=0.0)
    if h_ref is None:
        h_ref: float = click.prompt("--h_ref not provided, please enter a number, or Enter for default", default=0.0)
    if resolution is None:
        resolution: float = click.prompt(
            "--resolution not provided, please enter a number, or Enter for default", default=0.05
        )
    if window_size is None:
        window_size: int = click.prompt(
            "--window_size not provided, please enter a number, or Enter for default", default=10
        )
    if shapefile is not None:
        if dst is None:
            dst, crs_gcps = cli_utils.read_shape(shapefile)
            # validate if amount of points is logical
            dst = cli_utils.validate_dst(dst)
        else:
            logger.warning(
                f"Shapefile {shapefile} not used, because --dst was provided explicitly and overrules the use of "
                f"--shapefile."
            )
    if dst is not None:
        logger.info("Destination points found and validated")
    else:
        raise click.UsageError(
            "No destination control points for found, either provide a list of points with --dst or provide a "
            "shapefile with --shapefile"
        )
    if not src:
        logger.warning(
            "No source control points provided. No problem, you can interactively click them in your objective"
        )
        if click.confirm("Do you want to continue and provide source points interactively?", default=True):
            src, camera_matrix, dist_coeffs = cli_utils.get_gcps_interactive(
                videofile,
                dst,
                crs=crs,
                crs_gcps=crs_gcps,
                frame_sample=frame_sample,
                focal_length=focal_length,
                k1=k1,
                k2=k2,
                lens_position=lens_position,
                rotation=rotation,
                logger=logger,
            )
            if len(src) != len(dst):
                raise click.UsageError(f"You have not provided enough source points {len(src)}/{len(dst)} available")
    if crs is None and crs_gcps is not None:
        raise click.UsageError(f"--crs is None while --crs_gcps is {crs_gcps}, please supply --crs.")
    gcps = {"src": src, "dst": dst, "z_0": z_0, "h_ref": h_ref, "crs": crs_gcps}
    if not corners:
        logger.warning(
            "No corner points for projection provided. No problem, you can interactively click them in your objective"
        )
        if click.confirm("Do you want to continue and provide corners interactively?", default=True):
            corners = cli_utils.get_corners_interactive(
                videofile,
                gcps,
                crs=crs,
                crs_gcps=crs_gcps,
                frame_sample=frame_sample,
                camera_matrix=camera_matrix,
                dist_coeffs=dist_coeffs,
                rotation=rotation,
                logger=logger,
            )
    if stabilize:
        stabilize_pol = cli_utils.get_stabilize_pol(
            videofile,
            frame_sample=frame_sample,
            rotation=rotation,
            logger=logger,
        )
    else:
        stabilize_pol = None
    service.camera_config(
        video_file=videofile,
        cam_config_file=output,
        gcps=gcps,
        crs=crs,
        frame_sample=frame_sample,
        resolution=resolution,
        window_size=window_size,
        lens_position=lens_position,
        corners=corners,
        camera_matrix=camera_matrix,
        dist_coeffs=dist_coeffs,
        stabilize=stabilize_pol,
        rotation=rotation,
    )
    logger.info(f"Camera configuration created and stored in {output}")


# VELOCIMETRY
@cli.command(short_help="Estimate velocimetry")
@click.argument(
    "OUTPUT",
    type=click.Path(resolve_path=True, dir_okay=True, file_okay=False),
    callback=cli_utils.validate_dir,
    required=True,
)
@video_opt
@click.option(
    "-r",
    "--recipe",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False, file_okay=True),
    help="Options file (*.yml)",
    callback=cli_utils.parse_recipe,
    required=True,
)
@click.option(
    "-c",
    "--cameraconfig",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False, file_okay=True),
    help="Camera config file (*.json)",
    callback=cli_utils.parse_camconfig,
    required=True,
)
@click.option("-p", "--prefix", type=str, default="", help="Prefix for produced output files")
@click.option(
    "-h",
    "--h_a",
    type=float,
    required=False,
    help="Water level in local vertical datum (e.g. staff or pressure gauge) belonging to video.",
)
@click.option(
    "--cross",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False, file_okay=True),
    help="Cross section file (*.geojson). This will be used for discharge estimation if the `transect` "
    " section is provided in your recipe.",
    callback=cli_utils.parse_cross_section_gdf,
    required=False,
)
@click.option(
    "--cross_wl",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False, file_okay=True),
    help="Cross section file (*.geojson). If you provide this, you may add water level retrieval settings to the"
    " recipe in the section `water_level`. For more information see"
    " [PyORC docs](https://localdevices.github.io/pyorc/user-guide/cross_section/index.html)",
    callback=cli_utils.parse_cross_section_gdf,
    required=False,
)
@click.option(
    "-u",
    "--update",
    is_flag=True,
    default=False,
    help="Only update requested output files with changed inputs or if not present on file system",
)
@click.option(
    "--lowmem",
    is_flag=True,
    default=False,
    help="Reduce memory consumption. Computation will be slower",
)
@verbose_opt
@click.pass_context
def velocimetry(ctx, output, videofile, recipe, cameraconfig, prefix, h_a, cross, cross_wl, update, lowmem, verbose):
    """CLI subcommand for velocimetry."""
    log_level = max(10, 20 - 10 * verbose)
    logger = log.setuplog("velocimetry", os.path.abspath("pyorc.log"), append=False, log_level=log_level)
    logger.info(f"Preparing your velocimetry result in {output}")
    # load in recipe and camera config
    service.velocity_flow(
        recipe=recipe,
        videofile=videofile,
        cameraconfig=cameraconfig,
        h_a=h_a,
        cross=cross,
        cross_wl=cross_wl,
        prefix=prefix,
        output=output,
        update=update,
        concurrency=not (lowmem),
        logger=logger,
    )
    pass


# ensure help texts span a reasonable space
# velocimetry(max_content_width=120)

if __name__ == "__main__":
    # if getattr(sys, 'frozen', False):
    cli(sys.argv[1:])
